#-*- encoding: utf-8 -*-
#möp
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42):
# <dkratzert@gmx.de> wrote this file. As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some day,
# and you think this stuff is worth it, you can buy me a beer in return.
# Daniel Kratzert
# ----------------------------------------------------------------------------
#


import os
import re
import sys

import misc
from constants import sep_line

try:
    from argparse import RawTextHelpFormatter, Namespace
except ImportError:
    print('\n*** Your Python version is incompatible with DSR! \nPlease use Python 2.7 or above. ***\n')
    sys.exit()
from argparse import ArgumentParser, SUPPRESS





class OptionsParser():
    """
    This class uses the ArgumentParser module to parse the command line options.
    """
    def __init__(self, progname):
        self.progname = progname
        # search characters allowed: a-z A-Z 0-9 _ - , () {} [] ' " + * | = .
        self.allowed_chars = re.compile(r'^[\w\-,()\[\]{}\'\"+*|=.]+$')
        self.parser = ArgumentParser(prog='dsr', formatter_class=RawTextHelpFormatter,
                                     description='{}\nDisordered Structure Refinement (DSR)\n'.format(progname)
                                     + '\nExample DSR .res file command line:'
                                     + '\nREM DSR PUT/REPLACE "Fragment" WITH C1 C2 C3 ON Q1 Q2 Q3 PART 1 OCC -21 ='
                                     + '\n  RESI DFIX\n'
                                     + sep_line+'\n'
                                     + '   PUT:     Just put the fragment source atoms here.\n'
                                     + '   REPLACE: Replace atoms of PART 0 in 1.3 A distance around target atoms.\n'
                                     + sep_line
                                     , usage=SUPPRESS
                                     )
        self.parser.add_argument("-r", dest="res_file", metavar='"res file"', nargs='+',
                                 help="res file with DSR command", default='')
        self.parser.add_argument("-re", dest="external_restr", metavar='"res file"', nargs='+',
                                 help="res file with DSR command (write restraints to external file)", default=False)
        self.parser.add_argument("-e", dest="export_fragment", metavar='"fragment"',
                                 help="export fragment as .res/.png file", default='')
        self.parser.add_argument("-c", dest="export_clip", metavar='"fragment"',
                                 help="export fragment to clipboard", default='')
        self.parser.add_argument("-t", dest="invert", action='store_true',
                                 help="inverts the current fragment", default=False)
        self.parser.add_argument("-i", dest="import_grade", metavar='"tgz file"',
                                 help="import a fragment from GRADE (needs .tgz file)", default='')
        self.parser.add_argument("-l", dest="list_db", action="store_true",
                                 help="list names of all database entries", default=False)
        self.parser.add_argument("-s", dest="search_string", metavar='"string"', nargs='+',
                                 help="search the database for a name", default='')
        self.parser.add_argument("-g", dest="rigid_group", help="keep group rigid (no restraints)",
                                 action="store_true", default=False)
        self.parser.add_argument("-u", dest="selfupdate", help="Update DSR to the most current version",
                                 action="store_true", default=False)
        self.parser.add_argument("-ea", dest="export_all", action='store_true',
                                 help=SUPPRESS, default=False)
        self.parser.add_argument("-lc", dest="list_db_csv", action='store_true',
                                 help=SUPPRESS, default=False)
        self.parser.add_argument("-x", dest="search_extern", nargs='+',
                                 help=SUPPRESS, default=False)
        self.parser.add_argument("-ah", dest="head_for_gui",
                                 help=SUPPRESS, default='')
        # with nargs='+' it accepts space in path and returns a list:
        self.parser.add_argument("-shx", dest="shelxl_ex", nargs='+',
                                 help=SUPPRESS, default=False)
        self.parser.add_argument("-n", dest="no_refine", action="store_true",
                                 help="do not refine after fragment transfer", default=False)
        self.parser.add_argument("-target", dest="target", nargs='+', type=float,
                                 help=SUPPRESS, default=[])
        self._options = self.parser.parse_args()

    def error(self):
        self.parser.print_help()
        sys.exit()

    @property
    def res_file(self):
        if self._options.res_file:
            rpath = r' '.join(self._options.res_file)
        else:
            return False
        try:
            rpath = os.path.normpath(rpath)
        except Exception:
            rpath = None
        return rpath.strip()

    @property
    def external_restr(self):
        if self._options.external_restr:
            erpath = r' '.join(self._options.external_restr)
        else:
            return False
        try:
            erpath = os.path.normpath(erpath)
        except Exception:
            erpath = None
        return erpath

    @property
    def no_refine(self):
        return self._options.no_refine

    @property
    def rigid_group(self):
        return self._options.rigid_group

    @property
    def export_fragment(self) -> bool:
        return self._options.export_fragment

    @property
    def export_clip(self) -> str:
        return self._options.export_clip.strip()

    @property
    def export_all(self):
        return self._options.export_all

    @property
    def list_db(self):
        return self._options.list_db

    @property
    def list_db_csv(self):
        return self._options.list_db_csv

    @property
    def import_grade(self):
        return self._options.import_grade

    @property
    def selfupdate(self):
        return self._options.selfupdate

    @property
    def head_for_gui(self):
        frag = False
        if self._options.head_for_gui:
            frag = self._options.head_for_gui.lower().strip()
        return frag

    @property
    def invert(self) -> bool:
        return self._options.invert

    @property
    def target_coords(self):
        target = self._options.target
        if target and len(target) % 3 > 0:
            print("*** Number of target coordinates have to be triplets. [x y z x y z ...] ***")
            sys.exit()
        return target

    @property
    def shelxl_ex(self):
        """
        Option to define the path of the shelxl executable
        """
        spath = ''
        if self._options.shelxl_ex:
            spath = r' '.join(self._options.shelxl_ex)
        else:
            return False
        if not os.access(spath, os.X_OK):
            try:
                spath = misc.which(spath)
                return spath[0]
            except IndexError:
                return ''
        return spath

    @property
    def search_string(self):
        if not self._options.search_string:
            return None
        if not self.allowed_chars.match(''.join(self._options.search_string).strip()):
            print('*** Characters not allowed for searching. ***')
            sys.exit()
        else:
            return ''.join(self._options.search_string).strip()

    @property
    def search_extern(self):
        if not self._options.search_extern:
            return None
        if not self.allowed_chars.match(''.join(self._options.search_extern).strip()):
            print('*** Characters not allowed for searching. ***')
            sys.exit()
        else:
            return ''.join(self._options.search_extern)

    @property
    def all_options(self) -> Namespace:
        return self._options

    def __repr__(self):
        return self._options.__str__()


if __name__ == '__main__':
    optparse = OptionsParser(200)
    print(optparse)
    optparse.parser.print_help()





# -*- encoding: utf-8 -*-
# 
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42):
# <dkratzert@gmx.de> wrote this file. As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some day,
# and you think this stuff is worth it, you can buy me a beer in return.
# Daniel Kratzert
# ----------------------------------------------------------------------------
#


import shutil
from math import cos, sqrt, radians, sin

import os
import random
import re
import string
from pathlib import Path
from typing import List

import mpmath as mpm
from constants import isoatomstr

alphabet = string.ascii_uppercase


def write_file(flist, name):
    """
    Writes the content of list to name.
    :param flist: list
    :param name:  string
    """
    with open(name, 'w') as ofile:
        for line in flist:  # modified reslist
            ofile.write("%s" % line)  # write the new file


def extract_tarfile(file, targetdir):
    """
    Extracts .tar.gz "file" to "targetdir"
    :param file: .tar.gz file
    :param targetdir: target directory
    :return: bool for sucess
    """
    import tarfile
    try:
        with tarfile.open(file) as tarobj:
            tarobj.extractall(path=targetdir)
    except tarfile.ReadError as e:
        print('*** Could not extract tarfile ***')
        print('***', e, '***')
        return False
    return True


def join_floats(float_list, places=3, ):
    """
    >>> l = [1.23456789123456, 1, 2, 3, 3.45, 5.6543, 1,3456]
    >>> join_floats(l)
    '1.235 1.000 2.000 3.000 3.450 5.654 1.000 3456.000'
    """
    str_floats = " ".join(format(i, "{}.{}f".format(places + 2, places)) for i in float_list)
    return str_floats


def touch(fname, times=None):
    """
    creates an empty file like unix touch command
    """
    with open(fname, 'a'):
        os.utime(fname, times)


def check_file_exist(filename):
    """
    Check if a file exists and has some content.
    A file zize above 0 returns True.
    A non-existing file returns False.
    A file size of 0 retrurns 'zero'.

    # keep these paths for the unit tests!
    >>> check_file_exist('./tests/p21c.res')
    True
    >>> check_file_exist('foo.bar')
    File "foo.bar" not found!
    False
    >>> check_file_exist('./tests/empty.txt')
    'zero'
    >>> check_file_exist('src/dsr_shelx/misc.py')
    True
    """
    status = False
    if os.path.isfile(filename):
        filesize = int(os.stat(str(filename)).st_size)
    else:
        print('File "{}" not found!'.format(filename))
        return False
    if isinstance(filesize, int) and filesize > 0:
        status = True
    if isinstance(filesize, int) and filesize == 0:
        status = 'zero'
    return status


def walkdir(rootdir, include="", exclude=None):
    """
    Returns a list of files in all subdirectories with full path.
    :param rootdir: base path from which walk should start
    :param include: list of file endings to include only e.g. ['.py', '.res']
    :return: list of files

    >>> walkdir("../setup/modpath.iss")
    ['../setup/modpath.iss']
    >>> walkdir("../setup/modpath.iss", exclude=['.iss'])
    []
    """
    if exclude is None:
        exclude = [""]
    results = []
    if not os.path.isdir(rootdir):
        if os.path.splitext(rootdir)[1] in exclude:
            return []
        return [rootdir]
    for root, subFolders, files in os.walk(rootdir):
        for file in files:
            fullfilepath = os.path.join(root, file)
            if exclude:
                if os.path.splitext(fullfilepath)[1] in exclude:
                    continue
            if include:
                if os.path.splitext(fullfilepath)[1] in include:
                    results.append(os.path.normpath(fullfilepath).replace('\\', '/'))
            else:
                results.append(os.path.normpath(fullfilepath).replace('\\', '/'))
    return results


def pairwise(iterable):
    """
     s -> (s0,s1), (s2,s3), (s4, s5), ...

     >>> liste = ['C1', 'C2', 'C2', 'C3', 'C4', 'C5', 'C5', 'C6']
     >>> pairwise(liste)
     [('C1', 'C2'), ('C2', 'C3'), ('C4', 'C5'), ('C5', 'C6')]
     """
    a = iter(iterable)
    return list(zip(a, a))


def mean(values):
    """
    returns mean value of a list of numbers

    >>> mean([1, 2, 3, 4, 1, 2, 3, 4])
    2.5
    >>> round(mean([1, 2, 3, 4, 1, 2, 3, 4.1, 1000000]), 4)
    111113.3444
    """
    mean = sum(values) / float(len(values))
    return mean


def median(nums):
    """
    calculates the median of a list of numbers
    >>> median([2])
    2
    >>> median([1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4])
    2.5
    >>> median([1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4.1, 1000000])
    3
    >>> median([])
    Traceback (most recent call last):
    ...
    ValueError: Need a non-empty iterable
    """
    ls = sorted(nums)
    n = len(ls)
    if n == 0:
        raise ValueError("Need a non-empty iterable")
    # for uneven list length:
    elif n % 2 == 1:
        # // is floordiv:
        return ls[n // 2]
    else:
        return sum(ls[int(int(n) / 2 - 1):int(int(n) / 2 + 1)]) / 2.0


def std_dev(data):
    """
    returns standard deviation of values rounded to pl decimal places
    S = sqrt( (sum(x-xm)^2) / n-1 )
    xm = sum(x)/n
    >>> l1 = [1.334, 1.322, 1.345, 1.451, 1.000, 1.434, 1.321, 1.322]
    >>> l2 = [1.234, 1.222, 1.345, 1.451, 2.500, 1.234, 1.321, 1.222]
    >>> round(std_dev(l1), 8)
    0.13797871
    >>> round(std_dev(l2), 8)
    0.43536797
    >>> median(l1)
    1.328
    >>> mean(l1)
    1.316125
    """
    if len(data) == 0:
        return 0
    K = data[0]
    n = 0
    Sum = 0
    Sum_sqr = 0
    for x in data:
        n += 1
        Sum += x - K
        Sum_sqr += (x - K) * (x - K)
    variance = (Sum_sqr - (Sum * Sum) / n) / (n - 1)
    # use n instead of (n-1) if want to compute the exact variance of the given data
    # use (n-1) if data are samples of a larger population
    return sqrt(variance)


def nalimov_test(data):
    """
    returns a index list of outliers base on the Nalimov test for data.
    Modified implementation of:
    "R. Kaiser, G. Gottschalk, Elementare Tests zur Beurteilung von Messdaten
    Bibliographisches Institut, Mannheim 1972." 
    
    >>> data = [1.120, 1.234, 1.224, 1.469, 1.145, 1.222, 1.123, 1.223, 1.2654, 1.221, 1.215]
    >>> nalimov_test(data)
    [3]
    """
    # q-values for degrees of freedom:
    f = {1 : 1.409, 2: 1.645, 3: 1.757, 4: 1.814, 5: 1.848, 6: 1.870, 7: 1.885, 8: 1.895,
         9 : 1.903, 10: 1.910, 11: 1.916, 12: 1.920, 13: 1.923, 14: 1.926, 15: 1.928,
         16: 1.931, 17: 1.933, 18: 1.935, 19: 1.936, 20: 1.937, 30: 1.945}
    fact = sqrt(float(len(data)) / (len(data) - 1))
    fval = len(data) - 2
    if fval < 2:
        return []
    outliers = []
    if fval in f:
        # less strict than the original:
        q_crit = f[fval]
    else:
        q_crit = 1.95
    for num, i in enumerate(data):
        q = abs(((i - median(data)) / std_dev(data)) * fact)
        if q > q_crit:
            outliers.append(num)
    return outliers


def flatten(lis):
    """
    Given a list, possibly nested to any level, return it flattened.
    From: http://code.activestate.com/recipes/578948-flattening-an-arbitrarily-nested-list-in-python/

    >>> flatten([['wer', 234, 'brdt5'], ['dfg'], [[21, 34,5], ['fhg', 4]]])
    ['wer', 234, 'brdt5', 'dfg', 21, 34, 5, 'fhg', 4]
    """
    new_lis = []
    for item in lis:
        if type(item) == type([]):
            new_lis.extend(flatten(item))
        else:
            new_lis.append(item)
    return new_lis


def sortedlistdir(directory):
    """
    returns a sorted list of files in directory directory.
    :param directory: directory
    :type directory: string
    """
    try:
        dirlist = os.listdir(directory)
    except(IOError, OSError):
        return False
    dirlist.sort()
    return dirlist


def find_line_of_residue(reslist, resinumber):
    """
    Returns the line number where residue n appears in the reslist.
    :param reslist: res file as list like:
                    ['C1 1 -0.00146 0.26814 0.06351 11.00 0.05',
                    'RESI 4 BENZ',
                    'C2 1 -1.13341 -0.23247 -0.90730 11.00 0.05',]
    :type reslist: list of lists
    :param resinumber: residue number
    :type resinumber: string
    :return n: integer
    :return line: string
    >>> find_line_of_residue(['REM TEST', 'C1 1 -0.00146 0.26814 0.06351 11.00 0.05', \
        'RESI 4 BENZ', 'C2 1 -1.13341 -0.23247 -0.90730 11.00 0.05'], "4")
    [2, 'RESI 4 BENZ']
    """
    for n, line in enumerate(reslist):
        if line.upper().startswith('RESI'):
            if line.split()[1] == str(resinumber):
                return [n, line]


def ll_to_string(inputlist):
    """
    converts list of list to string with four whitespaces between each list element
    >>> ll_to_string([['REM', 'TEST', 'C1'], ['1', '-0.00146', '0.26814', '0.06351']])
    'REM   TEST   C1\\n1   -0.00146   0.26814   0.06351'
    """
    inputlist = [list(map(str, i)) for i in inputlist]
    newlist = []
    for i in inputlist:
        newlist.append('   '.join(i).rstrip())
    string = '\n'.join(newlist)
    return string


def multiline_test(line):
    """
    test if the current line is a multiline with "=" at the end
    :param line: 'O1 3 -0.01453 1.66590 0.10966 11.00 0.05 ='
    :type line: string
    >>> line = 'C1    1    0.278062    0.552051    0.832431    11.00000    0.02895    0.02285 ='
    >>> multiline_test(line)
    True
    >>> line = 'C1    1    0.278062    0.552051    0.832431    11.00000    0.05 '
    >>> multiline_test(line)
    False
    """
    line = line.rpartition('=')  # partition the line in before, sep, after
    line = ''.join(line[0:2])  # use all including separator
    line = line.rstrip()  # strip spaces
    if line.endswith('='):
        return True
    else:
        return False


def find_line(inputlist, regex, start=None):
    """
    returns the index number of the line where regex is found in the inputlist
    if stop is true, stop searching with first line found
    :param inputlist: list of strings
    :type inputlist: list
    :param regex: regular expression to search
    :type regex: string
    :param start: line number where to start the search
    :param start: start searching at line start
    :type start: string or int
    >>> inp = ['Hallo blub', 'foo bar blub', '123', '1 blub 2 3 4']
    >>> find_line(inp, '.*blub.*')
    0
    >>> find_line(inp, 'nonono')
    False
    >>> inp = [['foo'],['bar']]
    >>> find_line(inp, '.*blub.*') #doctest: +REPORT_NDIFF +NORMALIZE_WHITESPACE +ELLIPSIS
    Traceback (most recent call last):
        ...
    TypeError: expected string or ...
    """
    if start:
        start = int(start)
        inputlist_slice = inputlist[start:]
        for i, string in enumerate(inputlist_slice, start):
            if re.match(regex, string, re.IGNORECASE):
                return i  # returns the index number if regex found
    else:
        for i, string in enumerate(inputlist):
            if re.match(regex, string, re.IGNORECASE):
                return i  # returns the index number if regex found
    return False  # returns False if no regex found (xt solution has no fvar)


def remove_line(reslist, linenum, rem=False, remove=False, frontspace=False):
    """
    removes a single line from the res file with tree different methods.
    The default is a space character in front of the line (frontspace).
    This removes the line in the next refinement cycle. "rem" writes rem
    in front of the line and "remove" clears the line.
    :param reslist: .res file list
    :param linenum: integer, line number
    :param rem:     True/False, activate comment with 'REM' in front
    :param remove:  True/False, remove the line
    :param frontspace: True/False, activate removing with a front space
    """
    line = reslist[linenum]
    if rem:  # comment out with 'rem ' in front
        reslist[linenum] = 'rem ' + line
        if multiline_test(line):
            reslist[linenum + 1] = 'rem ' + reslist[linenum + 1]
    elif remove:  # really delete the line "linenum"
        if multiline_test(line):
            reslist[linenum] = ''
            reslist[linenum + 1] = ''
        else:
            reslist[linenum] = ''
    if frontspace:  # only put a space in front
        reslist[linenum] = ' ' + line
        if multiline_test(line):
            reslist[linenum + 1] = ' ' + reslist[linenum + 1]
    return reslist


def find_multi_lines(inputlist, regex):
    """
    returns the index number of all lines where regex is found in the inputlist
    ! this method is case insensitive !
    >>> input = ['Hallo blub', 'foo bar blub', '123', '1 blub 2 3 4']
    >>> find_multi_lines(input, '.*blub.*')
    [0, 1, 3]
    >>> find_multi_lines(input, 'blabla')
    []
    >>> inp = [['foo'],['bar']]
    >>> find_multi_lines(inp, r'.*blub.*') # doctest: +NORMALIZE_WHITESPACE +REPORT_NDIFF +ELLIPSIS
    Traceback (most recent call last):
        ...
    TypeError: expected string or ...
    """
    reg = re.compile(regex, re.IGNORECASE)
    foundlist = []
    for i, string in enumerate(inputlist):
        if reg.match(string):
            foundlist.append(i)  # returns the index number if regex found
    return foundlist


def remove_file(filename, exit_dsr=False, terminate=False):
    """
    removes the file "filename" from disk
    program exits when exit is true
    platon gets terminated if terminate is true

    >>> remove_file('foobar')
    """
    if os.path.isfile(filename):
        try:
            os.remove(filename)
        except(IOError, OSError):
            print('can not delete {}'.format(filename))
            # print 'unable to cleanup ins {} files!'.format(file)
            if terminate:
                pgrogname = terminate
                pgrogname.terminate()
            if exit_dsr:
                sys.exit()


def copy_file(source, target):
    """
    Copy a file from source to target. Source can be a single file or
    a directory. Target can be a single file or a directory.
    :param source: list or string
    :param target: string
    """
    target_path = os.path.dirname(target)
    source_file = os.path.basename(source)
    listcopy = False
    if isinstance(source, (list, tuple)):
        listcopy = True
    if not os.path.exists(target_path) and target_path != '':
        try:
            os.makedirs(target_path)
        except(IOError, OSError):
            print('Unable to create directory {}.'.format(target_path))
    try:
        if listcopy:
            for filen in source:
                shutil.copy(filen, target)
        else:
            shutil.copy(source, target)
    except IOError as e:
        print('Unable to copy {}.'.format(source_file))
        print(e)


def make_directory(dirpath):
    """
    create a directory with all subdirs from the last existing path
    :param dirpath: string
    """
    try:
        os.makedirs(dirpath)
    except(IOError, OSError):
        print('Unable to create directory {}.'.format(dirpath))
        return False
    return True


def wrap_headlines(dbhead, width=75):
    """
    wraps lines of a restraint header to prevent too long lines in
    SHELXL. wrapping is done with = at the end of a line and ' ' at
    start of the next line
    :param dbhead: header with restraints
    :param width: wrap after width characters
    >>> line = ['foo bar this is text to wrap. blah bub']
    >>> wrap_headlines(line, 15)
    ['foo bar this is =\\n   text to wrap. =\\n   blah bub\\n']
    >>> wrap_headlines(['SADI C1 C2 C3 C4'], 10)
    ['SADI C1 C2 =\\n   C3 C4\\n']
    """
    import textwrap
    for num, line in enumerate(dbhead):
        line = textwrap.wrap(line, width, subsequent_indent='  ')
        if len(line) > 1:
            newline = []
            for n, l in enumerate(line):
                if n < len(line) - 1:
                    l += ' =\n'
                newline.append(l)
            dbhead[num] = ' '.join(newline)
    for num, line in enumerate(dbhead):
        line = ' '.join(line.strip().split(' '))
        dbhead[num] = line + '\n'
    return dbhead


def wrap_stringlist(strlist, width=75):
    """
    Wraps the text lines of a list to width characters.
    >>> wrap_stringlist(['REM foo bar baz foo bar baz blah blub', 'REM 2foo bar baz foo bar baz blah blub'], 36)
    ['REM foo bar baz foo bar baz blah\\nREM blub\\n', 'REM 2foo bar baz foo bar baz blah\\nREM blub\\n']
    """
    import textwrap
    wrapped = []
    for num, line in enumerate(strlist):
        wrapped.append('\n'.join(textwrap.wrap(line, width, subsequent_indent='REM ')) + '\n')
    return wrapped


def wrap_text(inText, maxlen=70, subsequent_indent='=\n'):
    """
    Text wrapper without need for textwrap package.
    >>> wrap_text('SADI 0.02 C1A C2A C2A C3A C3A C4A C4A C5A C5A C6A', maxlen=30)
    'SADI 0.02 C1A C2A C2A C3A C3A =\\n C4A C4A C5A C5A C6A'
    """
    line_list = []
    wrapped = []
    inText_list = inText.strip().split()
    for n, word in enumerate(inText_list):
        testline = ' '.join(line_list)
        if len(word + testline + subsequent_indent) >= maxlen:
            if n < len(inText_list) - 1:
                line_list.append(word + ' ' + subsequent_indent)
            else:
                line_list.append(word)
            wrapped.append(' '.join(line_list).strip(' '))
            line_list = []
        else:
            line_list.append(word)
    if line_list:
        wrapped.append(' '.join(line_list).strip(' '))
    return ' '.join(wrapped)


def unwrap_head_lines(headlines):
    """
    if a line is wrapped like "SADI C1 C2 =\n", "  C3 C4" or "SADI C1 C2=\n", "  C3 C4"
    this function returns "SADI C1 C2 C3 C4"
    :type headlines: list
    :param headlines: list of strings from a res file
    >>> unwrap_head_lines(["SADI C1 C2 =\\n", "  C3 C4"])
    ['SADI C1 C2 C3 C4']
    >>> unwrap_head_lines(['foo bar this is =\\n   text to wrap. =\\n   blah bub\\n'])
    ['foo bar this is text to wrap. blah bub']
    >>> unwrap_head_lines(wrap_headlines(["REM name: ser\\nREM HFIX 23 C2 C3 C4  C6 C8 C10\\nSADI bar"]))
    ['REM name: ser\\nREM HFIX 23 C2 C3 C4  C6 C8 C10\\nSADI bar\\n']
    """
    import constants
    tmp = ''
    # it is faster with this loop:
    eq = False
    for line in headlines:
        if '=' in line:
            eq = True
            break
    if not eq:
        return headlines
    for line in headlines:
        line = line.strip(' \r\n=').replace('=', ' ')
        tmp = tmp + ' ' + line
    spline = tmp.split()
    last = ''
    for n, i in enumerate(spline):
        command = i[:4].upper()
        # leave out 'REM COMMAND':
        if command in constants.SHX_CARDS and last != 'REM':
            spline[n] = '\n' + spline[n]
        last = command
    new_head = ' '.join(spline).strip().split('\n')
    new_head = [i.rstrip(' ') for i in new_head]
    return new_head


def which(name, flags=os.X_OK, exts=None):
    r"""
    Search PATH for executable files with the given name.

    On MS-Windows the only flag that has any meaning is os.F_OK. Any other
    flags will be ignored.

    which('dsr.bat')
    ['d:\\\Programme\\\DSR\\\dsr.bat']
    """
    if exts is None:
        exts = ['.exe', '.EXE', '.bat']
    result = []
    # exts = filter(None, os.environ.get('PATHEXT', '').split(os.pathsep))
    path = os.getenv('PATH', None)
    if path is None:
        return []
    for p in os.getenv('PATH', '').split(os.pathsep):
        p = os.path.join(p, name)
        if os.access(p, flags):
            result.append(p)
        for e in exts:
            pext = p + e
            if os.access(pext, flags):
                result.append(pext)
    return result


def remove_partsymbol(atom):
    """
    strips the part symbol like C1_4b from an atom name
    :param atom: 'C1_4b'
    :type atom: string

    >>> remove_partsymbol('C2_4b')
    'C2_4'
    >>> remove_partsymbol('C22_b')
    'C22'
    >>> remove_partsymbol('C_5')
    'C_5'
    >>> remove_partsymbol('SAME/SADI')
    'SAME/SADI'
    >>> remove_partsymbol('C22_4^b')
    'C22_4'
    >>> remove_partsymbol('C23^b')
    'C23'
    >>> remove_partsymbol('C24_0^b')
    'C24'
    >>> remove_partsymbol('C25_0b')
    'C25'
    """
    if '_' in atom or '^' in atom:
        if '^' in atom:
            # since SHELXL 2016/5, residue and part are devided by "^"
            name = atom.split('^')[0]
            if name.split('_')[-1] == '0':
                return name.split('_')[0]
            else:
                return name
        else:
            presuff = atom.split('_')
            prefix, suffix = presuff[0], presuff[-1].strip(string.ascii_letters)
        if not suffix:
            return prefix
        else:
            if suffix == '0':
                atom = prefix
            else:
                atom = prefix + '_' + suffix
    return atom


def id_generator(size=6, chars=string.ascii_uppercase + string.digits):
    """
    returns a random ID like 'L5J74W'
    :param size: length of the string
    :type size: integer
    :param chars: characters used for the ID
    :type chars: string

    >>> id_generator(1, 'a')
    'a'
    """
    return ''.join(random.choice(chars) for _ in range(size))


def shift(seq, n):
    """
    left-shift a sliceable object by n
    :param seq: sequence to shift
    :type seq: string or list
    :param n: shift length
    :type n: int
    >>> shift('hello world', 3)
    'lo worldhel'
    >>> shift(['sdfg', 'dsfg', '111', '222'], 1)
    ['dsfg', '111', '222', 'sdfg']
    """
    n %= len(seq)
    return seq[n:] + seq[:n]


def atomic_distance(p1, p2, cell, shortest_dist=False):
    """
    p1 and p2 are x, y , z coordinates as list ['x', 'y', 'z']
    cell are the cell parameters as list: ['a', 'b', 'c', 'alpha', 'beta', 'gamma']

    Returns the distance between the two points (Atoms). If shortest_dist is True, the
    shortest distance ignoring translation is computed.

    >>> cell = (10.5086, 20.9035, 20.5072, 90, 94.13, 90)
    >>> coord1 = (-0.186843,   0.282708,   0.526803)
    >>> coord2 = (-0.155278,   0.264593,   0.600644)
    >>> atomic_distance(coord1, coord2, cell)
    1.5729229943265979
    >>> cell = [14.0452, 14.0452, 12.2008, 90.000, 90.000, 90.000]
    >>> atomic_distance([-0.184235, 0.955143, -0.181577], [0.437437, 1.162938, 0.182932], cell)
    10.224259725884533
    >>> atomic_distance([0.455143, 1.184235, 0.181577], [0.437437, 1.162938, 0.182932], cell, shortest_dist=True)
    0.38934604708396064
    """
    a, b, c = cell[:3]
    al = radians(cell[3])
    be = radians(cell[4])
    ga = radians(cell[5])
    if shortest_dist:
        x1, y1, z1 = [x + 99.5 for x in p1]
        x2, y2, z2 = p2
        dx = (x1 - x2) % 1 - 0.5
        dy = (y1 - y2) % 1 - 0.5
        dz = (z1 - z2) % 1 - 0.5
    else:
        x1, y1, z1 = p1
        x2, y2, z2 = p2
        dx = (x1 - x2)
        dy = (y1 - y2)
        dz = (z1 - z2)
    dsq = (a * dx) ** 2 + (b * dy) ** 2 + (c * dz) ** 2 + 2 * b * c * cos(al) * dy * dz + \
          2 * dx * dz * a * c * cos(be) + 2 * dx * dy * a * b * cos(ga)
    return sqrt(dsq)


def frac_to_cart(frac_coord, cell):
    """
    Converts fractional coordinates to cartesian coodinates
    :param frac_coord: [float, float, float]
    :param cell:       [float, float, float, float, float, float]
    >>> cell = (10.5086, 20.9035, 20.5072, 90, 94.13, 90)
    >>> coord1 = (-0.186843,   0.282708,   0.526803)
    >>> print(frac_to_cart(coord1, cell))
    [-2.741505423999065, 5.909586678000002, 10.775200700893734]
    """
    a, b, c, alpha, beta, gamma = cell
    x, y, z = frac_coord
    alpha = radians(alpha)
    beta = radians(beta)
    gamma = radians(gamma)
    try:
        cosastar = (cos(beta) * cos(gamma) - cos(alpha)) / (sin(beta) * sin(gamma))
        sinastar = sqrt(1 - cosastar ** 2)
    except ValueError:
        print("*** Malformed unit cell parameters found! Please correct the database entry. ***")
        sys.exit()
    Xc = a * x + (b * cos(gamma)) * y + (c * cos(beta)) * z
    Yc = 0 + (b * sin(gamma)) * y + (-c * sin(beta) * cosastar) * z
    Zc = 0 + 0 + (c * sin(beta) * sinastar) * z
    return [Xc, Yc, Zc]


class A(object):
    """
    orthogonalization matrix
    e.g. converts fractional coordinates to cartesian coodinates

    >>> cell = (10.5086, 20.9035, 20.5072, 90, 94.13, 90)
    >>> coord = (-0.186843,   0.282708,   0.526803)
    >>> A = A(cell).orthogonal_matrix
    >>> print(mpm.nstr(A*mpm.matrix(coord)))
    [-2.74151]
    [ 5.90959]
    [ 10.7752]
    >>> cartcoord = mpm.matrix([['-2.74150542399906'], ['5.909586678'], ['10.7752007008937']])
    >>> print(mpm.nstr(A**-1*cartcoord))
    [-0.186843]
    [ 0.282708]
    [ 0.526803]
    """

    def __init__(self, cell):
        self.a, self.b, self.c, alpha, beta, gamma = cell
        self.V = vol_unitcell(self.a, self.b, self.c, alpha, beta, gamma)
        self.alpha = radians(alpha)
        self.beta = radians(beta)
        self.gamma = radians(gamma)

    @property
    def orthogonal_matrix(self) -> mpm.matrix:
        """
        Converts von fractional to cartesian.
        Invert the matrix to do the opposite.
        """
        Am = mpm.matrix([[self.a, self.b * cos(self.gamma), self.c * cos(self.beta)],
                         [0, self.b * sin(self.gamma),
                          (self.c * (cos(self.alpha) - cos(self.beta) * cos(self.gamma)) / sin(self.gamma))],
                         [0, 0, self.V / (self.a * self.b * sin(self.gamma))]])
        return Am


def cart_to_frac(cart_coord, cell):
    """
    converts cartesian coordinates to fractional coordinates
    :param cart_coord: [float, float, float]
    :param cell:       [float, float, float, float, float, float]
    >>> cell = (10.5086, 20.9035, 20.5072, 90, 94.13, 90)
    >>> coords = [-2.74150542399906, 5.909586678, 10.7752007008937]
    >>> cart_to_frac(coords, cell)
    [-0.186843, 0.282708, 0.526803]
    """
    a, b, c, alpha, beta, gamma = cell
    X, Y, Z = cart_coord
    alpha = radians(alpha)
    beta = radians(beta)
    gamma = radians(gamma)
    cosastar = (cos(beta) * cos(gamma) - cos(alpha)) / (sin(beta) * sin(gamma))
    sinastar = sqrt(1 - cosastar ** 2)
    z = Z / (c * sin(beta) * sinastar)
    y = (Y - (-c * sin(beta) * cosastar) * z) / (b * sin(gamma))
    x = (X - (b * cos(gamma)) * y - (c * cos(beta)) * z) / a
    return [round(x, 8), round(y, 8), round(z, 8)]


def zero(m, n):
    """
    Create zero matrix of dimension m,n
    :param m: integer
    :param n: integer

    >>> zero(5, 3)
    [[0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0]]
    """
    new_matrix = [[0 for row in range(n)] for col in range(m)]  # @UnusedVariable
    return new_matrix


def determinante(a):
    """
    return determinant of 3x3 matrix
    Deprecated, use mpmath instead!!!

    >>> m1 = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]
    >>> determinante(m1)
    8
    """
    return (a[0][0] * (a[1][1] * a[2][2] - a[2][1] * a[1][2])
            - a[1][0] * (a[0][1] * a[2][2] - a[2][1] * a[0][2])
            + a[2][0] * (a[0][1] * a[1][2] - a[1][1] * a[0][2]))


def subtract_vect(a, b):
    """
    subtract vector b from vector a
    Deprecated, use mpmath instead!!!
    :param a: [float, float, float]
    :param b: [float, float, float]

    >>> subtract_vect([1, 2, 3], [3, 2, 2])
    (-2, 0, 1)
    """
    return (a[0] - b[0],
            a[1] - b[1],
            a[2] - b[2])


def matrix_minus_vect(m, v):
    """
    >>> source = [[-0.01453, 1.6659, 0.10966], [-0.00146, 0.26814, 0.06351], [-0.27813, -0.21605, 1.52795]]
    >>> cent = [-0.09804,     0.57266333 , 0.56704]
    >>> matrix_minus_vect(source, cent) # DOCTEST: +NORMALIZE_WHITESPACE +ELLIPSIS
    [[0.08351, 1.09323667, -0.45738], [0.09658, -0.30452333000000004, -0.50353], [-0.18008999999999997, -0.78871333, 0.9609099999999999]]
    >>> matrix_minus_vect([[1,1,1],[2,2,2],[3,3,3]], [1,1,1])
    [[0, 0, 0], [1, 1, 1], [2, 2, 2]]
    """
    result = []
    for coords in m:
        result.append([coords[0] - v[0], coords[1] - v[1], coords[2] - v[2]])
    return result


def matrix_plus_vect(m, v):
    """
    >>> source = [[-0.01453, 1.6659, 0.10966], [-0.00146, 0.26814, 0.06351], [-0.27813, -0.21605, 1.52795]]
    >>> cent = [-0.09804,     0.57266333 , 0.56704]
    >>> matrix_plus_vect(source, cent)
    [[-0.11257, 2.23856333, 0.6767], [-0.0995, 0.84080333, 0.6305499999999999], [-0.37617, 0.35661333000000006, 2.09499]]
    >>> matrix_minus_vect([[1,1,1],[2,2,2],[3,3,3]], [1,1,1])
    [[0, 0, 0], [1, 1, 1], [2, 2, 2]]
    """
    result = []
    for coords in m:
        result.append([coords[0] + v[0], coords[1] + v[1], coords[2] + v[2]])
    return result


def transpose(a):
    """
    transposes a matrix

    >>> m = [[1, 2, 3], [1, 2, 3], [1, 2, 3]]
    >>> transpose(m)
    [(1, 1, 1), (2, 2, 2), (3, 3, 3)]
    """
    return list(zip(*a))


def norm_vec(a):
    """
    returns a normalized vector

    >>> norm_vec([1, 2, 1])
    (0.4082482904638631, 0.8164965809277261, 0.4082482904638631)
    """
    l = sqrt(a[0] ** 2 + a[1] ** 2 + a[2] ** 2)
    return a[0] / l, a[1] / l, a[2] / l


def vol_tetrahedron(a, b, c, d, cell=None):
    """
    returns the volume of a terahedron spanned by four points:
    e.g. A = (3, 2, 1), B = (1, 2, 4), C = (4, 0, 3), D = (1, 1, 7)

            |u1 u2 u3|
    v = 1/6*|v1 v2 v3|
            |w1 w2 w3|

    AB = (1-3, 2-2, 4-1) = (-2, 0, 3)
    AC = ...
    AD = ...

    V = 1/6[u,v,w]

              |-2,  0, 3|
    [u,v,w] = | 1, -2, 2| = 24-3-12 = 5
              |-2, -1, 6|

    V = 1/6*5

    >>> cell = (10.5086, 20.9035, 20.5072, 90, 94.13, 90)
    >>> a = (0.838817,   0.484526,   0.190081) # a ist um 0.01 ausgelenkt
    >>> b = (0.875251,   0.478410,   0.256955)
    >>> c = (0.789290,   0.456520,   0.301616)
    >>> d = (0.674054,   0.430194,   0.280727)
    >>> print('volume of Benzene ring atoms:')
    volume of Benzene ring atoms:
    >>> print(round(vol_tetrahedron(a, b, c, d, cell), 7))
    0.0633528
    """
    A = [float(i) for i in a]
    B = [float(i) for i in b]
    C = [float(i) for i in c]
    D = [float(i) for i in d]
    if cell:
        A = frac_to_cart(a, cell)
        B = frac_to_cart(b, cell)
        C = frac_to_cart(c, cell)
        D = frac_to_cart(d, cell)
    AB = subtract_vect(A, B)
    AC = subtract_vect(A, C)
    AD = subtract_vect(A, D)
    D = determinante([AB, AC, AD])
    volume = abs((D / 6))
    return volume


def vol_unitcell(a, b, c, al, be, ga):
    """
    calculates the volume of a unit cell
    >>> v = vol_unitcell(2, 2, 2, 90, 90, 90)
    >>> print(v)
    8.0
    """
    ca, cb, cg = cos(radians(al)), cos(radians(be)), cos(radians(ga))
    v = a * b * c * sqrt(1 + 2 * ca * cb * cg - ca ** 2 - cb ** 2 - cg ** 2)
    return v


def dice_coefficient(a, b, case_insens=True):
    """
    :type a: str
    :type b: str
    :type case_insens: bool
    dice coefficient 2nt/na + nb.
    https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Dice%27s_coefficient#Python
    >>> dice_coefficient('hallo', 'holla')
    0.25
    >>> dice_coefficient('Banze', 'Benzene')
    0.444444
    >>> dice_coefficient('halo', 'Haaallo')
    0.75
    >>> dice_coefficient('hallo', 'Haaallo')
    0.888889
    >>> dice_coefficient('hallo', 'Hallo')
    1.0
    >>> dice_coefficient('aaa', 'BBBBB')
    0.0
    """
    if case_insens:
        a = a.lower()
        b = b.lower()
    if not len(a) or not len(b):
        return 0.0
    if len(a) == 1:
        a = a + '.'
    if len(b) == 1:
        b = b + '.'
    a_bigram_list = []
    for i in range(len(a) - 1):
        a_bigram_list.append(a[i:i + 2])
    b_bigram_list = []
    for i in range(len(b) - 1):
        b_bigram_list.append(b[i:i + 2])
    a_bigrams = set(a_bigram_list)
    b_bigrams = set(b_bigram_list)
    overlap = len(a_bigrams & b_bigrams)
    dice_coeff = overlap * 2.0 / (len(a_bigrams) + len(b_bigrams))
    return round(dice_coeff, 6)


def dice_coefficient2(a, b, case_insens=True):
    """ 
    :type a: str
    :type b: str
    :type case_insens: bool
    duplicate bigrams in a word should be counted distinctly
    (per discussion), otherwise 'AA' and 'AAAA' would have a 
    dice coefficient of 1...
    https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Dice%27s_coefficient#Python

    This implementation is reverse. 1 means not hit, 0 means best match
    >>> dice_coefficient2('hallo', 'holla')
    0.5
    >>> dice_coefficient2('Banze', 'Benzene')
    0.8
    >>> dice_coefficient2('halo', 'Haaallo')
    1.333333
    >>> dice_coefficient2('hallo', 'Haaallo')
    1.6
    >>> dice_coefficient2('hallo', 'Hallo')
    2.0
    >>> dice_coefficient2('aaa', 'BBBBB')
    0.0
    >>> dice_coefficient2('', '')
    0.0
    """
    if case_insens:
        a = a.lower()
        b = b.lower()
    if not len(a) or not len(b):
        return 0.0
    # quick case for true duplicates
    if a == b:
        return 2.0
    # if a != b, and a or b are single chars, then they can't possibly match
    if len(a) == 1 or len(b) == 1:
        return 0.0
    # use python list comprehension, preferred over list.append()
    a_bigram_list = [a[i:i + 2] for i in range(len(a) - 1)]
    b_bigram_list = [b[i:i + 2] for i in range(len(b) - 1)]
    a_bigram_list.sort()
    b_bigram_list.sort()
    # assignments to save function calls
    lena = len(a_bigram_list)
    lenb = len(b_bigram_list)
    # initialize match counters
    matches = i = j = 0
    while i < lena and j < lenb:
        if a_bigram_list[i] == b_bigram_list[j]:
            matches += 2
            i += 1
            j += 1
        elif a_bigram_list[i] < b_bigram_list[j]:
            i += 1
        else:
            j += 1
    score = float(2 * matches) / float(lena + lenb)
    return round(score, 6)


def longest_common_substring(s1, s2):
    """
    returns the longest common substring of two strings
    :param s1: a string
    :type s1: str
    :param s2: a second string
    :type s2: str

    >>> longest_common_substring('hello world how is foo bar?', 'hello daniel how is foo in the world?')
    ' how is foo '
    """
    m = [[0] * (1 + len(s2)) for i in range(1 + len(s1))]  # @UnusedVariable
    longest, x_longest = 0, 0
    for x in range(1, 1 + len(s1)):
        for y in range(1, 1 + len(s2)):
            if s1[x - 1] == s2[y - 1]:
                m[x][y] = m[x - 1][y - 1] + 1
                if m[x][y] > longest:
                    longest = m[x][y]
                    x_longest = x
            else:
                m[x][y] = 0
    return s1[x_longest - longest: x_longest]


def levenshtein(s1, s2):
    """
    >>> levenshtein('hallo', 'holla')
    2
    >>> dice_coefficient('hallo', 'holla')
    0.25
    """
    s1 = s1.lower()
    s2 = s2.lower()
    if len(s1) < len(s2):
        return levenshtein(s2, s1)
    if len(s2) == 0:
        return len(s1)
    previous_row = list(range(len(s2) + 1))
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            # j+1 instead of j since previous_row and current_row are one character longer:
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1  # than s2
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row
    return previous_row[-1]


def distance(x1, y1, z1, x2, y2, z2, round_out=0):
    """
    distance between two points in space for orthogonal axes.
    >>> distance(1, 1, 1, 2, 2, 2, 4)
    1.7321
    >>> distance(1, 0, 0, 2, 0, 0, 4)
    1.0
    """
    from math import sqrt
    d = sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2 + (z1 - z2) ** 2)
    if round_out:
        return round(d, round_out)
    else:
        return d


def coord_to_shx_atom(coordinates):
    """
    Transformes a list of coordinates to a list of shelxl atom strings.
    """
    strlist = []
    sfac_num = 1
    for num, coord in enumerate(coordinates):
        at = "AX" + str(num)
        s = isoatomstr.format(at, sfac_num, coord[0], coord[1], coord[2], 11.0000, 0.03)
        strlist.append(s)
    return strlist


def ufrac_to_ucart(_A, cell, uvals):
    U11, U22, U33, U23, U13, U12 = uvals
    U21 = U12
    U32 = U23
    U31 = U13
    Uij = mpm.matrix([[U11, U12, U13], [U21, U22, U23], [U31, U32, U33]])
    a, b, c, alpha, beta, gamma = cell
    V = vol_unitcell(*cell)
    # calculate reciprocal lattice vectors:
    astar = (b * c * sin(radians(alpha))) / V
    bstar = (c * a * sin(radians(beta))) / V
    cstar = (a * b * sin(radians(gamma))) / V
    # matrix with the reciprocal lattice vectors:
    N = mpm.matrix([[astar, 0, 0],
                    [0, bstar, 0],
                    [0, 0, cstar]])
    # Finally transform Uij values from fractional to cartesian axis system:
    Ucart = _A * N * Uij * N.T * _A.T
    return Ucart


def almost_equal(a, b, places=3):
    """
    Returns True or False if the number a and b are are equal inside the
    decimal places "places".
    :param a: a real number
    :type a: int/float
    :param b: a real number
    :type b: int/float
    :param places: number of decimal places
    :type places: int

    >>> almost_equal(1.0001, 1.0005)
    True
    >>> almost_equal(1.1, 1.0005)
    False
    >>> almost_equal(2, 1)
    False
    """
    return round(abs(a - b), places) == 0


def file_len(fname):
    """
    Returns the number of lines in a text file.

    >>> file_len('tests/p21c.hkl')
    42976
    """
    i = 0
    with open(fname) as f:
        for i, l in enumerate(f, 1):
            pass
    return i


def get_overlapped_chunks(ring, size):
    """
    returns a list of chunks of size 'size' which overlap with one field.
    If the last chunk is smaller than size, the last 'size' chunks are returned as last chunk.
    "size" has to be larger than 3 to get reasonable results.
    >>> l = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
    >>> get_overlapped_chunks(l, 4)
    [[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10], [10, 11, 12, 13], [12, 13, 14, 15]]
    >>> get_overlapped_chunks(l, 3)  # Not working with that chunk size
    [[13, 14, 15], [13, 14, 15], [13, 14, 15], [13, 14, 15], [13, 14, 15]]
    >>> get_overlapped_chunks(l, 5)
    [[1, 2, 3, 4, 5], [4, 5, 6, 7, 8], [7, 8, 9, 10, 11], [10, 11, 12, 13, 14], [11, 12, 13, 14, 15]]
    """
    chunks = []
    for i in range(0, len(ring) - size + 3, 3):
        chunk = ring[i:i + size]
        if len(chunk) < 4:
            chunk = ring[-size:]
        chunks.append(sorted(chunk))
    return chunks


def chunks(l, n):
    """
    returns successive n-sized chunks from l.

    >>> l = [1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 'a', 'b', 'c', 'd', 'e', 'f']
    >>> chunks(l, 5)
    [[1, 2, 3, 4, 5], [6, 7, 8, 9, 0], ['a', 'b', 'c', 'd', 'e'], ['f']]
    >>> chunks(l, 1)
    [[1], [2], [3], [4], [5], [6], [7], [8], [9], [0], ['a'], ['b'], ['c'], ['d'], ['e'], ['f']]
    >>> chunks(l, 50)
    [[1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 'a', 'b', 'c', 'd', 'e', 'f']]
    """
    out = []
    for i in range(0, len(l), n):
        out.append(l[i:i + n])
    return out


def read_cif(lines: List[str], debug=False):
    """
    Open a Crystallographic Information File (*.cif) file and store all entries in a key:value dictionary
     Looped values are stored as lists under a single key entry
     All values are stored as strings
    E.G.
      crys=readcif('somefile.cif')
      crys['_cell_length_a'] = '2.835(2)'

    crys[key] = value
    available keys are give by crys.keys()

    Modified version of the CIF parser in https://github.com/DanPorter/Dans_Diffraction
    """
    cifvals = {}

    # Read file line by line, converting the cif file values to a python dict
    n = 0
    while n < len(lines):
        # Convert line to columns
        vals = lines[n].strip().split()

        # skip empty lines
        if len(vals) == 0:
            n += 1
            continue

        # Search for stored value lines
        if vals[0][0] == '_':
            if len(vals) == 1:
                # Record next lines that are not keys as string
                if lines[n + 1][0] == ';':
                    n += 1
                strarg = []
                while n + 1 < len(lines) and (len(lines[n + 1]) == 0 or lines[n + 1][0].strip() not in ['_', ';']):
                    strarg += [lines[n + 1].strip('\'"')]
                    n += 1
                cifvals[vals[0]] = '\n'.join(strarg)
                chk = 'a'
            else:
                cifvals[vals[0]] = ' '.join(vals[1:]).strip(' \'"\n')
                chk = 'b'
            n += 1
            if debug:
                print('%5d %s %s = %s' % (n, chk, vals[0], cifvals[vals[0]]))

        # Search for loops
        elif vals[0] == 'loop_':
            n += 1
            loopvals = []
            # Step 1: Assign loop columns
            # looped columns are given by "_column_name"
            while n < len(lines) and len(lines[n].strip()) > 0 and lines[n].strip()[0] == '_':
                loopvals += [lines[n].split()[0]]
                cifvals[loopvals[-1]] = []
                n += 1

            # Step 2: Assign data to columns
            # loops until line has less segments than columns
            while n < len(lines):
                # cols = lines[n].split()
                # this fixes error on symmetry arguments having spaces
                # this will only work if the last argument in the loop is split by spaces (in quotes)
                # cols = cols[:len(loopvals) - 1] + [''.join(cols[len(loopvals) - 1:])]
                cols = [col for col in re.split("( |\\\".*?\\\"|'.*?')", lines[n]) if col.strip()]
                if len(cols) != len(loopvals):
                    break
                if cols[0][0] == '_' or cols[0] == 'loop_':
                    break  # catches error if loop is only 1 iteration
                if cols[0][0] == '#':
                    n += 1;
                    continue  # catches comented out lines
                if len(loopvals) == 1:
                    cifvals[loopvals[0]] += [lines[n].strip(' \"\'\n')]
                else:
                    for c, ll in enumerate(loopvals):
                        cifvals[ll] += [cols[c]]
                n += 1
            if debug:
                for ll in loopvals:
                    print('%5d L %s = %s' % (n, ll, str(cifvals[ll])))
        else:
            # Skip anything else
            if debug:
                print('%5d SKIPPED: %s' % (n, lines[n]))
            n += 1

    # Replace '.' in keys - fix bug from isodistort cif files
    # e.g. '_space_group_symop_magn_operation.xyz'
    current_keys = list(cifvals.keys())
    for key in current_keys:
        if '.' in key:
            newkey = key.replace('.', '_')
            cifvals[newkey] = cifvals[key]
    return cifvals


if __name__ == '__main__':
    import sys
    import doctest

    failed, attempted = doctest.testmod()  # verbose=True)
    if failed == 0:
        print('passed all {} tests!'.format(attempted))



# -*- encoding: utf-8 -*-
# möp
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42):
# <dkratzert@gmx.de> wrote this file. As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some day,
# and you think this stuff is worth it, you can buy me a beer in return.
# Daniel Kratzert
# ----------------------------------------------------------------------------
#


import os
import sys

import misc


def filename_wo_ending(resfilename):
    """
    returns the input file name without ending
    :param resfilename: string like 'p21c.res'
    """
    if not resfilename:
        return False
    file_ext = os.path.splitext(resfilename)
    basefile = file_ext[0]
    if not file_ext[1] in ['.res', '']:
        print("Please give a .res file as argument.")
        sys.exit(0)
    return str(basefile)


class ResList():
    """
    Reads and writes the res-file as list
    """

    def __init__(self, res_file_name):
        """
        :param res_file_name: string like 'p21c.res'
        """
        self.__res_file_name = res_file_name
        self.__basefile = filename_wo_ending(self.__res_file_name)

    def get_res_list(self):
        """
        read the res-file and return it as list
        """
        filename = self.__basefile + '.res'
        reslist = []
        try:
            with open(filename, 'r') as f:
                for line in f:
                    reslist.append(line)
        except IOError:
            print('Unable to read {} file.'.format(filename))
            sys.exit()
        return reslist

    def write_resfile(self, reslist, ending):
        """
        Writes the content of reslist to basefile+ending.
        Ending should be '.ins' or '.res'
        :param reslist: SHELXL .res file as list
        :param ending:  string, file ending like '.res'
        """
        try:
            nfile = open(self.__basefile + ending, 'w')  # open the ins file
        except IOError:
            print('Unable to write .res file! Check directory write permissions.')
            sys.exit(-1)
        for line in reslist:  # modified reslist
            if line == ' ':
                line += ' \n'  # prevent space character coming in front of an atom accidentally
            nfile.write("%s" % line)  # write the new file
        nfile.close()


class ResListEdit():
    """
    Class to add, remove and list lines of the resfile
    """

    def __init__(self, reslist, find_atoms):
        """

        :param reslist: SHELXL .res file as list
        :param find_atoms: FindAtoms() object
        """
        self._reslist = reslist
        self._find_atoms = find_atoms
        self.fvarlines = self.find_fvarlines()
        self.cell = self._cell_from_res()

    def get_cell(self):
        """
        returns cell parameters of the res file
        """
        return self.cell
 
    def _cell_from_res(self):
        cell = []
        for line in self._reslist:
            if line.startswith('CELL'):
                cell = line.split()[2:8]
                cell = [float(i) for i in cell]
                return cell
        return cell

    def add_line(self, linenum, new_line):
        """
        adds a single line of string to the res file
        checks are missing which make sure the data structure of new_line is correct!
        :param linenum:  integer, line number to add string
        :param new_line: new string to insert like 'foo bar'
        """
        linenum = linenum - 1
        self._reslist.insert(linenum, new_line + '\n')
        return self._reslist

    def remove_line(self, linenum, rem=False, remove=False, frontspace=False):
        """
        removes a single line from the res file with tree different methods.
        The default is a space character in front of the line (frontspace).
        This removes the line in the next refinement cycle. "rem" writes rem
        in front of the line and "remove" clears the line.
        :param linenum: integer, line number
        :param rem:     True/False, activate comment with 'REM' in front
        :param remove:  True/False, remove the line
        :param frontspace: True/False, activate removing with a front space
        """
        line = self._reslist[linenum]
        if rem:  # comment out with 'rem ' in front
            self._reslist[linenum] = 'rem ' + line
            if misc.multiline_test(line):
                self._reslist[linenum + 1] = 'rem ' + self._reslist[linenum + 1]
        elif remove:  # really delete the line "linenum"
            if misc.multiline_test(line):
                self._reslist[linenum] = ''
                self._reslist[linenum + 1] = ''
            else:
                self._reslist[linenum] = ''
        if frontspace:  # only put a space in front
            self._reslist[linenum] = ' ' + line
            if misc.multiline_test(line):
                self._reslist[linenum + 1] = ' ' + self._reslist[linenum + 1]
        return self._reslist

    def list_lines(self, startline, endline):
        """
        returns the lines between startline and endline as list
        :param startline: integer
        :param endline:   integer
        """
        lines = []
        try:
            start = int(startline)
            end = int(endline)
        except(NameError, SyntaxError):
            return False
        for line, i in enumerate(self._reslist):
            if line >= start:
                lines.append(i)
            if line == end:
                break
        return lines

    def find_fvarlines(self):
        """
        Finds the FVAR line or the first line with an atom.
        returns a list of FVAR occurences
        sys.exit() if no atom or FVAR found
        """
        first_atom = 0
        fvarlines = misc.find_multi_lines(self._reslist, r'^FVAR')
        if not fvarlines:  # There is no FVAR in the res file after SHELXS!
            for num, i in enumerate(self._reslist):
                if self._find_atoms.is_atom(i):
                    first_atom = num
                    break
            if not first_atom:
                print('\nNo atom or Q-peak found! Can not proceed...\n')
                sys.exit()
            fvarlines = []
            fvarlines.append(first_atom)
            self._reslist.insert(first_atom, ' \n')
            return fvarlines
        else:
            return fvarlines

    def insert_frag_fend_entry(self, dbatoms, cell, position):
        """
        Inserts the FRAG ... FEND entry in the res file.
        :param dbatoms:   list of atoms in the database entry
        :type dbatoms: list
        :param cell:  list with cell from the database entry
        :type cell: list
        :param position: line number where FVAR or the first atom is located
        :type position: int
        """
        dblines = []
        db = [list(map(str, i)) for i in dbatoms]
        for i in db:
            i[2] = '{:>10.6f}'.format(float(i[2]))
            i[3] = '{:>10.6f}'.format(float(i[3]))
            i[4] = '{:>10.6f}'.format(float(i[4]))
            dblines.append('    '.join(i).rstrip())
        dblines = '\n'.join(dblines)
        dblines = 'FRAG 17 {} {} {} {} {} {}'.format(*cell) + '\n' + dblines
        dblines = '\n  The following is from DSR:\n' + dblines
        dblines = dblines + '\nFEND\n\n'
        self._reslist.insert(position + 1, dblines)  # insert the db entry right after FVAR

    def get_fvarlist(self):
        fvar_list = []
        for line in self.fvarlines:
            fvar = self._reslist[line].split()
            if fvar:
                del fvar[0]
            fvar_list.extend(fvar)
        return fvar_list

    def set_free_variables(self, occupancynumber, fvalue='0.5'):
        """
        Inserts additional free variables according to the occ parameter
        This function starts at the end of parse_dsr_line() so we don't have
        to care about it anywhere else.
        :param occupancynumber: string, like '21.0'
        :type occupancynumber: string
        :param fvarlines:       list, list of line numbers where FVAR is located
                                      in the res file
        """
        fvar_list = self.get_fvarlist()  # free variables in the res file
        varlen = self.get_fvar_count()
        occupancynumber = occupancynumber.strip('-')
        # how many numbers do we have?:
        # the occupancynumber is split in the fvar part and the occupancy part:
        num = occupancynumber.split('.')
        fvar = int(num[0]) // 10  # e.g. 20.5 is fvar 2 and occupancy 0.5
        if fvar == 0:
            fvar = 1
        difference = (fvar - varlen)
        if fvar > 1:
            if difference > 0:
                for i in range(difference):
                    fvar_list.append(fvalue)  # if an fvar is missing, add a new one
            else:  # make sure the occupancy of the disoerder parts get not < 0:
                if len(fvar_list) - (fvar - 1) >= 0:  # make sure
                    fvar_value = fvar_list[fvar - 1]
                    if (float(occupancynumber) - (10 * int(fvar) + float(fvar_value))) < 0:
                        fvar_list[fvar - 1] = '0.5'
        fvar_list = [str(x) for x in fvar_list]
        lines = misc.chunks(fvar_list, 7)
        if len(fvar_list) != 0:
            for line in self.fvarlines:
                self.remove_line(line, remove=True)  # removes the old FVAR
        fvars = [' '.join(i) for i in lines]
        fvars = ['FVAR ' + i for i in fvars]
        self._reslist[self.fvarlines[0]] = ' \n'.join(fvars) + '\n'
        return fvars

    def get_fvar_count(self):
        """
        returns the last used free variable defined with FVAR
        """
        fvars = len(self.get_fvarlist())
        if not fvars:
            fvars = 0
        return fvars


# for testing
if __name__ == '__main__':
    pass



# -*- encoding: utf-8 -*-
"""
Created on 13.05.2015

@author: Daniel Kratzert
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42):
# <dkratzert@gmx.de> wrote this file. As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some day,
# and you think this stuff is worth it, you can buy me a beer in return.
# Daniel Kratzert
# ----------------------------------------------------------------------------
#
"""

import string
import sys

from math import radians, sqrt

import mpmath as mpm
from atomhandling import Elem_2_Sfac, NumberScheme
from elements import get_radius_from_element
from misc import atomic_distance, frac_to_cart, cart_to_frac, \
    id_generator, shift, remove_partsymbol, find_multi_lines, wrap_headlines, ufrac_to_ucart, A
from refine import ShelxlRefine
from resfile import ResList
from restraints import ListFile

# Y-Z-F1/F2/F3

dfixr_130 = ['DFIX 1.328 Z F1 Z F2 Z F3 ',
             'DFIX 2.125 F1 F2 F2 F3 F3 F1 ',
             'SADI 0.1   Y F1 Y F2 Y F3 ',
             'RIGU Y Z F1 F2 F3 ',
             'SIMU Y Z F1 F2 F3 ']

dfixr_120 = ['DFIX 1.328 Z F1 Z F2 Z F3  Z F4 Z F5 Z F6 ',
             'DFIX 2.125 F1 F2 F2 F3 F3 F1  F4 F5 F5 F6 F6 F4 ',
             'SADI 0.1   Y F1 Y F2 Y F3  Y F4 Y F5 Y F6 ',
             'RIGU Y Z F1 > F6',
             'SIMU Y Z F1 > F6']

dfixr_120_split = ['DFIX 1.328 ZA F1 ZA F2 ZA F3  ZB F4 ZB F5 ZB F6 ',
                   'DFIX 2.125 F1 F2 F2 F3 F3 F1  F4 F5 F5 F6 F6 F4 ',
                   'SADI 0.1   Y F1 Y F2 Y F3  Y F4 Y F5 Y F6 ',
                   'SADI Y ZA Y ZB',
                   'EADP ZA ZB',
                   'RIGU Y ZA ZB F1 > F6',
                   'SIMU Y ZA ZB F1 > F6']

dfixr_cf9 = ['SUMP 1 0.0001 1 {0} 1 {1} 1 {2}',
             'DFIX 1.328 Z F1 Z F2 Z F3  Z F4 Z F5 Z F6  Z F7 Z F8 Z F9 ',
             'DFIX 2.125 F1 F2 F2 F3 F3 F1  F4 F5 F5 F6 F6 F4  F7 F8 F8 F9 F9 F7 ',
             'SADI 0.1 Y F1 Y F2 Y F3  Y F4 Y F5 Y F6  Y F7 Y F8 Y F9 ',
             'RIGU Y Z F1 > F9',
             'SIMU Y Z F1 > F9']

sadir_130 = ['SADI 0.02 Z F1 Z F2 Z F3 ',
             'SADI 0.04 F1 F2 F2 F3 F3 F1 ',
             'SADI 0.1  Y F1 Y F2 Y F3 ',
             'RIGU Y Z F1 F2 F3 ',
             'SIMU Y Z F1 F2 F3 ']

sadir_120 = ['SADI 0.02 Z F1 Z F2 Z F3  Z F4 Z F5 Z F6 ',
             'SADI 0.04 F1 F2 F2 F3 F3 F1  F4 F5 F5 F6 F6 F4 ',
             'SADI 0.1  Y F1 Y F2 Y F3  Y F4 Y F5 Y F6 ',
             'RIGU Y Z F1 > F6',
             'SIMU Y Z F1 > F6']

sadir_120_split = ['SADI 0.02 ZA F1 ZA F2 ZA F3  ZB F4 ZB F5 ZB F6 ',
                   'SADI 0.04 F1 F2 F2 F3 F3 F1  F4 F5 F5 F6 F6 F4 ',
                   'SADI 0.1  Y F1 Y F2 Y F3  Y F4 Y F5 Y F6 ',
                   'SADI Y ZA Y ZB',
                   'EADP ZA ZB',
                   'RIGU Y ZA ZB F1 > F6',
                   'SIMU Y ZA ZB F1 > F6']

sadir_cf9 = ['SUMP 1 0.0001 1 {0} 1 {1} 1 {2}',
             'SADI 0.02 Z F1 Z F2 Z F3  Z F4 Z F5 Z F6  Z F7 Z F8 Z F9 ',
             'SADI 0.04 F1 F2 F2 F3 F3 F1  F4 F5 F5 F6 F6 F4  F7 F8 F8 F9 F9 F7 ',
             'SADI 0.1  Y F1 Y F2 Y F3  Y F4 Y F5 Y F6  Y F7 Y F8 Y F9 ',
             'RIGU Y Z F1 > F9',
             'SIMU Y Z F1 > F9']


class CF3(object):
    '''
    a class to create cf3 groups at terminal atoms
    '''

    def __init__(self, rle, fa, reslist, fragment, sfac_table, basefilename, dsrp, resi, res_file, options):
        """
        Constructor
        """
        self.options = options
        self.resi = resi
        self.rand_id = id_generator(size=7)
        self.rand_id_part_neg = id_generator(size=7)
        self.fa = fa
        self.rle = rle
        self.dsrp = dsrp
        self.fragment = fragment
        self.reslist = reslist
        self.res_file = res_file
        self.rl = ResList(res_file)
        self.e2s = Elem_2_Sfac(sfac_table)
        atoms = fa.atoms_as_residues()
        self.basefilename = basefilename
        atomlist = []
        # atomlist: ['C1', ['x', 'y', 'z'], linenumber, class, part, element,
        #                                           sfac_number, residue_num]
        for i in atoms:
            for y in atoms[i]:
                if y[0][0] == 'Q':
                    continue
                atomlist.append(y + [i])
        self.atomlist = atomlist
        self.cell = rle.get_cell()
        self.startm = '\nREM CF3 group made by DSR:\n'
        self.endm = 'REM End of CF3 group made by DSR\n'

    def find_bonded_fluorine(self, atom, extra_param=0.16, element='F'):
        """
        find fluorine atoms that are boneded to atom
        returns ['C1', ['x', 'y', 'z'], linenumber, class, part, element, sfac_number, residue_num]
        """
        found_atoms = []
        atcoord = self.fa.get_atomcoordinates([atom])
        cr = get_radius_from_element('C')
        fr = get_radius_from_element(element)
        for i in self.atomlist:
            if not i[5] == element:
                continue
            d = atomic_distance(i[1], atcoord[atom], self.cell)
            if d <= (fr + cr) + extra_param and d > (cr or fr):
                found_atoms.append(i)
        return found_atoms

    def delete_bound_fluorine(self, bound_atoms):
        """
        deletes fluorine atoms bound to atom
        :param bound_atoms:
        :type bound_atoms:
        """
        for i in bound_atoms:
            i = int(i[2])
            self.rle.remove_line(i, rem=False, remove=True, frontspace=False)

    def format_cf3_restraints(self, afix, restr, atom, fatoms, splitatoms=False):
        """
        replaces the dummy atom names in the restraint lists with the real names
        :param afix: string of the afix number
        :param restr: restraints
        :param atom: pivot atom
        :param fatoms: fluorine atoms
        """
        Y, Z = self.lf.get_bondvector(atom)
        Y = remove_partsymbol(Y)
        Z = remove_partsymbol(Z)
        replacelist = ()
        if afix == '130':
            F1, F2, F3 = fatoms
            replacelist = (('Z', Z), ('Y', Y), ('F1', F1), ('F2', F2), ('F3', F3))
        if afix == '120' and splitatoms:
            F1, F2, F3, F4, F5, F6 = fatoms
            ZA, ZB = splitatoms
            replacelist = (('ZA', ZA), ('ZB', ZB), ('Y', Y), ('F1', F1), ('F2', F2),
                           ('F3', F3), ('F4', F4), ('F5', F5), ('F6', F6))
        if afix == '120' and not splitatoms:
            F1, F2, F3, F4, F5, F6 = fatoms
            replacelist = (('Z', Z), ('Y', Y), ('F1', F1), ('F2', F2), ('F3', F3),
                           ('F4', F4), ('F5', F5), ('F6', F6))
        # replace dummy atoms in restraint list with real atom names:
        for old, new in replacelist:
            restr = [i.replace(old, new) for i in restr]
        restr = wrap_headlines(restr, 77)
        return restr

    def make_pivot_isotropic(self, linenumber: int):
        """
        make sure the pivot atom of a cf3 group is isotropic
        :param linenumber: line number (index) in self.reslist of the pivot atom
        :type linenumber: integer
        :return Uij values, coordinates: U and xyz values of the pivot atom as lists
        """
        atomline = self.reslist[linenumber].split()
        coords = [float(atomline[2]), float(atomline[3]), float(atomline[4])]
        if atomline[-1] == '=':
            nextline = self.reslist[linenumber + 1].split()
            try:
                U11, U22 = atomline[6], atomline[7]
                U33, U23, U13, U12 = nextline[0], nextline[1], nextline[2], nextline[3]
            except Exception:
                # In this case we have a U value missing
                print('*** Incomplete Uij values. Atom split not possible! ***')
                self.dsrp.split = False
                return [[], coords]
            self.reslist[linenumber] = '{:5.4s}{:4.2s}{:>10.8s} {:>10.8s} {:>10.8s}  {:8.6s}  0.04'.format(*atomline)
            self.reslist[linenumber + 1] = ''
            return [[float(U11), float(U22), float(U33), float(U23), float(U13), float(U12)], coords]
        else:
            # atom is already isotropic, nothing to do...
            if self.dsrp.split:
                print('Pivot atom is isotropic. Atom split not possible!')
                self.dsrp.split = False
            return [[], coords]

    def add_chars(self, atom, alphabet):
        """
        add chars to an atom name until the name is unique
        :param atom:
        """
        for char in alphabet:
            if atom + char not in [i[0] for i in self.atomlist]:
                del alphabet[0]
                return atom + char

    def prepare_cf3(self):
        """
        prepares some things before CF3 group is generated
        :return: list
        """
        targetatom = self.dsrp.target[0]
        if targetatom.startswith("Q"):
            print("*** Only carbon atoms allowed for CF3-groups! ***")
            sys.exit()
        if len(self.dsrp.target) > 1:
            print('Using only first target atom {}.'.format(targetatom))
        try:
            atomline = self.fa.get_atom_line_numbers([targetatom])[0]
        except IndexError:
            print("\n*** Atom {} not found ***\n".format(targetatom))
            sys.exit()
        if '_' in targetatom:
            print('\n*** Sorry, can not create a CF3 group inside a residue! '
                  '\nThis would damage the residue. ***')
            sys.exit()
        found = self.find_bonded_fluorine(targetatom)
        for i in found:
            print('Deleting {}_{} from {}'.format(i[0], i[7], targetatom))
        self.delete_bound_fluorine(found)
        return atomline

    def cf3(self, afix='130'):
        """
        create CF3 group on atom.
        Either define atom at startup or let dsrparser get the atom name.
        Y-Z-F1/F2/F3
        AFIX 130: CF3 group
        AFIX 120: CF6 group

        :param afix: afix number for the CF3 group
        :type afix: string
        """
        # the pivot atom of the CF3 group:
        atomline = self.prepare_cf3()
        afix = str(afix)
        splitat1 = ''
        splitat2 = ''
        axes = []
        targetatom = self.dsrp.target[0]
        # c_coords: coordinates of the pivot atom
        uvals, c_coords = self.make_pivot_isotropic(atomline)
        restr = ['']
        if afix == '130':
            print('Generating CF3-Group at {}.'.format(targetatom))
            restr = sadir_130
            if self.dsrp.dfix:
                restr = dfixr_130
        if afix == '120':
            print('Generating twofold disordered CF3-Group at {}.'.format(targetatom))
            restr = sadir_120
            if self.dsrp.split:
                restr = sadir_120_split
            if self.dsrp.dfix:
                restr = dfixr_120
                if self.dsrp.split:
                    restr = dfixr_120_split
        if afix == '120' and self.dsrp.split and uvals:
            num = NumberScheme(self.reslist, [targetatom], self.dsrp)
            if len(targetatom) < 4:
                # in this case it is possible to add a character
                alphabet = [i for i in string.ascii_uppercase]
                splitat1 = self.add_chars(targetatom, alphabet)
                splitat2 = self.add_chars(targetatom, alphabet)
            else:
                splitat1 = num.get_fragment_number_scheme()[0]
                splitat2 = num.get_fragment_number_scheme(extranames=[splitat1])[0]
            splitatoms = [splitat1, splitat2]
            axes = calc_ellipsoid_axes(c_coords, uvals, self.cell)
        else:
            splitatoms = False
        # The fluorine atoms are generated here:
        fatoms = self.make_afix(afixnum=int(afix), linenumber=atomline)
        if not fatoms:
            return False
        self.do_refine_cycle(self.rl, self.reslist)
        # this is essential
        self.reslist = self.rl.get_res_list()
        # this is the bond around the CF3 group rotates
        restr = self.format_cf3_restraints(afix, restr, targetatom, fatoms, splitatoms)
        # get position for the fluorine atoms and make sure the reslist is the newest:
        self.fa._reslist = self.reslist
        atomline = self.fa.get_atom_line_numbers([targetatom])[0]
        # add restraints to reslist:
        restr = ''.join(restr)
        if not self.dsrp.split:
            self.reslist[atomline] = self.reslist[atomline] + self.startm + restr
        else:
            self.reslist[atomline] = ''
            self.reslist[atomline + 1] = self.reslist[atomline] + self.startm + restr
        id_lines = find_multi_lines(self.reslist, r'.*{}'.format(self.rand_id))
        neg_part_id_lines = find_multi_lines(self.reslist, r'.*{}'.format(self.rand_id_part_neg))
        # replace dummy PART with real part definition and C-atom coords with split coords
        if self.dsrp.split and afix == '120':
            at1 = '{:<5s} {:<3} {:>9.6f}   {:>9.6f}   {:>9.6f}   {:>8.4f}     0.04\n' \
                .format(splitat1, self.e2s.elem_2_sfac('C'), axes[0][0],
                        axes[0][1], axes[0][2], float(self.dsrp.occupancy))
            at2 = '{:<5s} {:<3} {:>9.6f}   {:>9.6f}   {:>9.6f}   {:>8.4f}     0.04\n' \
                .format(splitat2, self.e2s.elem_2_sfac('C'), axes[1][0],
                        axes[1][1], axes[1][2], -float(self.dsrp.occupancy))
            dummy = ''
            splb = False
            # put all together to build up the atoms:
            for line, splatom in zip(id_lines, [at1, at2, dummy]):
                # restraints should never be placed in this reslist[line]:
                self.reslist[line] = ' '.join(self.reslist[line].split()[1:3]) + '\n'
                self.reslist[line] = self.reslist[line] + splatom
                # exchange the first three Fluorine atoms with coordinates from one
                # disorder direction and the last three with the other direction:
                for num, fline in enumerate(self.reslist[line + 1:line + 1 + 3]):
                    fline = fline.split()
                    if not splatom:
                        continue
                    [fsplit_a, fsplit_b] = calc_ellipsoid_axes(fline[2:5], uvals, self.cell)
                    resline = self.reslist[line + 1 + num].split()
                    resline[5] = float(resline[5])
                    resline[2:5] = fsplit_a
                    if splb:
                        resline[2:5] = fsplit_b
                    if num == 2 and not splb:
                        # switch to fsplit_b. I can not decide on num, because that goes two times
                        # from 0 to 2
                        splb = True
                    self.reslist[line + 1 + num] = '{:<5s} {:<3} {:>9.6f}   {:>9.6f}   {:>9.6f}   {:>8.4f}     0.04\n' \
                        .format(*resline)
        else:
            for line in id_lines:
                # replacing "REM PART 1 !QNGUYQ2" with "PART 1"
                # restraints should never be placed in this reslist[line]:
                self.reslist[line] = ' '.join(self.reslist[line].split()[1:3]) + '\n'
        # remove PART -1 around AFIX 130:
        for line in neg_part_id_lines:
            self.reslist[line] = ''
        # set refinement cycles back to 8
        shx = ShelxlRefine(self.reslist, self.basefilename, self.fa, self.options)
        shx.set_refinement_cycles('8')
        self.rl.write_resfile(self.reslist, '.res')
        return fatoms

    def cf9(self):
        """
        create disorderd CF3 group on three positions.
        """
        atomline = self.prepare_cf3()
        target_atom = self.dsrp.target[0]
        print('Generating threefold disordered CF3-Group at {}.'.format(target_atom))
        if self.dsrp.dfix:
            restr = dfixr_cf9
        else:
            restr = sadir_cf9
        # make a copy to find fluorine positions:
        ####################################################
        reslist_copy = self.reslist[:]
        self.make_pivot_isotropic(atomline)
        # Fluorine atoms are generated here:
        fatoms = self.make_afix(afixnum=130, linenumber=atomline, resioff=True)
        self.do_refine_cycle(self.rl, self.reslist)
        # this is the bond around the CF3 group rotates
        Y, Z = self.lf.get_bondvector(target_atom)
        Y = remove_partsymbol(Y)
        Z = remove_partsymbol(Z)
        numberedatoms = ['F1', 'F2', 'F3', 'F4', 'F5', 'F6', 'F7', 'F8', 'F9']
        resistr = ''
        resi0 = ''
        if self.dsrp.resiflag:
            F1, F2, F3, F4, F5, F6, F7, F8, F9 = numberedatoms
            resiclass = self.resi.get_residue_class
            resinum = self.resi.get_resinumber
            resistr = 'RESI ' + resiclass + ' ' + resinum
            resi0 = 'RESI 0\n'
        else:
            nums = NumberScheme(self.reslist, numberedatoms, self.dsrp)
            F1, F2, F3, F4, F5, F6, F7, F8, F9 = nums.get_fragment_number_scheme()
        start_f_coord = self.lf.get_single_coordinate(fatoms[0] + '^A')
        replacelist = (('Z', Z), ('Y', Y), ('F1', F1), ('F2', F2), ('F3', F3),
                       ('F4', F4), ('F5', F5), ('F6', F6), ('F7', F7),
                       ('F8', F8), ('F9', F9))
        for old, new in replacelist:
            restr = [i.replace(old, new) for i in restr]
        ########################################################
        # add restraints to reslist:
        # all fluorines are set, so we can get back to the original res file:
        self.reslist = reslist_copy
        self.rle._reslist = self.reslist
        # Turned out to be problematic with ShelXle:
        # if self.dsrp.occupancy:
        #    occ = self.dsrp.occupancy
        # else:
        occ = str((self.rle.get_fvar_count() + 1) * 10 + 1 + 20)
        fcount = self.rle.get_fvar_count()
        self.rle.set_free_variables(occ, '0.3')
        atomline = self.fa.get_atom_line_numbers([target_atom])[0]
        self.make_pivot_isotropic(atomline)
        atoms_cf9 = ['PART 1 {1}1',
                     F1 + '   {0}   {4}    11.00000    0.04',
                     F2 + '   {0}   {5}    11.00000    0.04',
                     F3 + '   {0}   {6}    11.00000    0.04',
                     'PART 2 {2}1',
                     F4 + '   {0}   {7}    11.00000    0.04',
                     F5 + '   {0}   {8}    11.00000    0.04',
                     F6 + '   {0}   {9}    11.00000    0.04',
                     'PART 3 {3}1',
                     F7 + '   {0}   {10}    11.00000    0.04',
                     F8 + '   {0}   {11}    11.00000    0.04',
                     F9 + '   {0}   {12}    11.00000    0.04',
                     'PART 0 ',
                     self.endm]
        restr = wrap_headlines(restr, 77)
        restr = ''.join(restr).format(fcount + 1, fcount + 2, fcount + 3)
        # place the restraints:
        self.reslist[atomline] = (self.reslist[atomline] + self.startm + restr)
        at1 = self.lf.get_single_coordinate(Y)
        at2 = self.lf.get_single_coordinate(Z)
        coords = []
        # rotate the fluorine coordinate around at1, at2 to get three triples:
        for delta in [0, 120, 240, 40, 160, 280, 80, 200, 320]:
            coord = self.rotate_atom_around_bond(start_f_coord, at1, at2, delta)
            coords.append('{:>10.6f} {:>10.6f} {:>10.6f}'.format(*coord))
        if self.dsrp.resiflag:
            atoms_cf9 = [resistr] + atoms_cf9 + [resi0]
        # join all together:
        atoms_cf9 = '\n'.join(atoms_cf9).format(self.e2s.elem_2_sfac('F'),
                                                fcount + 1, fcount + 2, fcount + 3, *coords)
        self.reslist[atomline] += atoms_cf9
        # have to do this here, because set_free_variables() works on different reslist:
        # self.reslist[self.rle.find_fvarlines()[0]] = ' \n'.join(fvar)+'\n'
        shx = ShelxlRefine(self.reslist, self.basefilename, self.fa, self.options)
        shx.set_refinement_cycles('8')
        self.rl.write_resfile(self.reslist, '.res')
        return fatoms

    def make_afix(self, afixnum, linenumber, resioff=False):
        """
        create an afix to build a CF3 or CH3 group
        :param resioff: bool
        :param linenumber:  line number in the reslist where fragment is placed
        :param afixnum: afix number
        :type afixnum: int
        """
        resistr = ''
        resi0 = ''
        # Turned out to be problematic with ShelXle:
        # if self.dsrp.occupancy:
        #    occ = self.dsrp.occupancy
        # else:
        numberscheme_130 = []
        numberscheme_120 = []
        occ = str((self.rle.get_fvar_count() + 1) * 10 + 1)
        if int(afixnum) == 120:
            self.dsrp.occupancy = occ
            self.rle.set_free_variables(occ, '0.5')
            num_120 = NumberScheme(self.reslist, ['F1', 'F2', 'F3', 'F4', 'F5', 'F6'], self.dsrp)
            # returns also the atom names if residue is active
            numberscheme_120 = num_120.get_fragment_number_scheme()
        if int(afixnum) == 130:
            num_130 = NumberScheme(self.reslist, ['F1', 'F2', 'F3'], self.dsrp)
            # returns also the atom names if residue is active
            numberscheme_130 = num_130.get_fragment_number_scheme()
        if self.dsrp.resiflag and not resioff:
            resiclass = self.resi.get_residue_class
            resinum = self.resi.get_resinumber
            resistr = '\nRESI ' + resiclass + ' ' + resinum
            resi0 = 'RESI 0\n'
            numberscheme_130 = ['F1', 'F2', 'F3']
            numberscheme_120 = ['F1', 'F2', 'F3', 'F4', 'F5', 'F6']
        sfac = self.e2s.elem_2_sfac('F')
        if int(afixnum) == 130:
            # CF3:
            afix_130 = [resistr,
                        'AFIX {0}\nPART -1 !{4}',  # AFIX 120 or 130
                        # 'REM AFIX made by DSR: {3}',
                        numberscheme_130[0] + ' {1}   0 0 0   11.0  0.04',
                        numberscheme_130[1] + ' {1}   0 0 0   11.0  0.04',
                        numberscheme_130[2] + ' {1}   0 0 0   11.0  0.04',
                        # 'REM end of AFIX by DSR {3}', # insert ID and later change it to the PART usw.
                        'PART 0 !{4}\nAFIX 0',
                        resi0]
            afix_130 = '\n'.join(afix_130)
            afix = afix_130
        elif int(afixnum) == 120:
            # CF6:
            # {4} is a second id behind the PART -1 to get rid of it later
            # We need a negative part here to prevent accidental bonding to the pivot atom
            afix_120 = [resistr,
                        'AFIX {0}\nPART -1 !{4}',
                        '\nREM PART 1 !{3}',
                        numberscheme_120[0] + ' {1}  0 0 0   {2}  0.04',
                        numberscheme_120[1] + ' {1}  0 0 0   {2}  0.04',
                        numberscheme_120[2] + ' {1}  0 0 0   {2}  0.04',
                        'REM PART 2 !{3}',
                        numberscheme_120[3] + ' {1}  0 0 0  -{2}  0.04',
                        numberscheme_120[4] + ' {1}  0 0 0  -{2}  0.04',
                        numberscheme_120[5] + ' {1}  0 0 0  -{2}  0.04',
                        'REM PART 0 !{3}',
                        'PART 0 !{4}\nAFIX 0',
                        resi0,
                        self.endm]
            afix_120 = '\n'.join(afix_120)
            afix = afix_120
        else:
            print('Only AFIX 130 and 120 are implemented yet.')
            return False
        # insert the afix (the rand_id is to recognize this line later):
        self.reslist[linenumber] += afix.format(afixnum, sfac, occ, self.rand_id, self.rand_id_part_neg)
        if str(afixnum) == '120':
            return numberscheme_120
        if str(afixnum) == '130':
            return numberscheme_130

    def do_refine_cycle(self, rl, reslist):
        """
        runs a shelxl cycle with L.S. 0
        :param rl: reslist object
        :param reslist: res file as list
        """
        shx = ShelxlRefine(reslist, self.basefilename, self.fa, self.options)
        acta_lines = shx.remove_acta_card()
        shx.set_refinement_cycles('0')
        rl.write_resfile(reslist, '.ins')
        shx.run_shelxl()
        self.lf = ListFile(self.basefilename)
        lst_file = self.lf.read_lst_file()
        shx.check_refinement_results(lst_file)
        rl = ResList(self.res_file)
        reslist = rl.get_res_list()
        shx.restore_acta_card(acta_lines)
        shx.set_refinement_cycles('8')
        rl.write_resfile(reslist, '.res')

    def rotate_atom_around_bond(self, ratom, at1, at2, delta=10):
        """
        R = T**-1*Rx**-1*Ry**-1*Rz*Ry*Rx*T

        1 translate object to origin
        2 rotate around x in xz plane
        3 rotate around y in z axis
        4 rotate aound z with delta angle
        5 transform back with inverse of 1-3

        v = P2 - P1
        |v| = sqrt(vx**2+vy**2+vz**2)
        u = v/|v| = (a, b, c), |u| = 1
        a = (x2-x1)/|v| b = (x2-y1)/|v| c = ...
        """
        ratom = frac_to_cart(ratom, self.cell)
        at1 = frac_to_cart(at1, self.cell)
        at2 = frac_to_cart(at2, self.cell)

        ratom = mpm.matrix(list(ratom) + [1.0])
        delta = radians(delta)
        x0, y0, z0 = at1
        T = mpm.matrix(((1, 0, 0, -x0),
                        (0, 1, 0, -y0),
                        (0, 0, 1, -z0),
                        (0, 0, 0, 1)))
        T1 = mpm.inverse(T)

        vx = at2[0] - at1[0]
        vy = at2[1] - at1[1]  # P2 - P1
        vz = at2[2] - at1[2]
        vnorm = sqrt(vx ** 2 + vy ** 2 + vz ** 2)
        a, b, c = vx / vnorm, vy / vnorm, vz / vnorm
        d = mpm.sqrt(b ** 2 + c ** 2)

        sina = b / d  # For rotation around alpha
        cosa = c / d  #

        Rxa = mpm.matrix(((1, 0, 0, 0),
                          (0, cosa, sina, 0),
                          (0, -sina, cosa, 0),
                          (0, 0, 0, 1)))
        '''
        Rya = mpm.matrix(((      cosa,          0,      -sina,  0),
                         (         0,          1,          0,  0),
                         (      sina,          0,       cosa,  0),
                         (         0,          0,          0,  1)))
        ''' '''
        Rza = mpm.matrix(((      cosa,       sina,          0,  0),
                         (     -sina,       cosa,          0,  0),
                         (         0,          0,          1,  0),
                         (         0,          0,          0,  1)))
        '''
        Rxa1 = mpm.inverse((Rxa))

        cosb = d  #
        sinb = -a  # rotation around beta

        Ryb = mpm.matrix(((cosb, 0, -sinb, 0),
                          (0, 1, 0, 0),
                          (sinb, 0, cosb, 0),
                          (0, 0, 0, 1)))

        Ryb1 = mpm.inverse((Ryb))

        sind = mpm.sin(delta)
        cosd = mpm.cos(delta)

        Rzd = mpm.matrix(((cosd, sind, 0, 0),
                          (-sind, cosd, 0, 0),
                          (0, 0, 1, 0),
                          (0, 0, 0, 1)))

        R = T1 * Rxa * Ryb * Rzd * Ryb1 * Rxa1 * T
        v = R * ratom
        v = cart_to_frac(v[:3], self.cell)
        v = [round(i, 6) for i in v]
        return v

    def make_cf3_thorus(self, atom=None):
        """
        This method was just an experiment. The CF9 method is sufficient.

        Creates a thorus of isotropic fluorine atoms around the central
        atom of a cf3 group. The occupancy is estimated from the
        residual density values in the lst file.

        :param atom: central atom of the cf3 group
        :type atom: string
        """
        coords = []
        if not atom:
            atom = self.dsrp.target[0]
        # returns the atom names of the fluorine atoms:
        fluorine_names = self.cf3()
        # self.do_refine_cycle(self.rl, self.reslist)
        print(fluorine_names[0], '##################')
        ratom = self.lf.get_single_coordinate(fluorine_names[0])
        print('ratom:', ratom)
        # ratom = self.get_coordinates_of_first_atom(fluorine_names)
        self.lf.read_lst_file()
        bondvec = self.lf.get_bondvector()
        print('bondvec:', bondvec)
        at1 = self.lf.get_single_coordinate(bondvec[0].upper())
        at2 = self.lf.get_single_coordinate(bondvec[1].upper())
        names = ['F{}'.format(i) for i in range(1, 25)]
        num_thor = NumberScheme(self.reslist, names, self.dsrp)
        names = num_thor.get_fragment_number_scheme()
        for delta in range(0, 360, 15):
            coord = self.rotate_atom_around_bond(ratom, at1, at2, delta)
            coords.append(coord)
        # print('PART 0')
        diffden = self.lf.get_difference_density(averaged=False)
        print(diffden)
        rotation = self.lf.get_degree_of_highest_peak()
        print(rotation)
        n = (rotation / 15) + 4  # +1
        diffden = shift(diffden, n)
        print(len(diffden))
        print('\n')
        print('')
        print('AFIX 6')
        # make restraints:
        sad12 = []
        sad13 = []
        ff = []
        flast = False
        for i, num, co, dif in zip(names, [i for i in range(1, 25)], coords, diffden):
            # print('PART {}'.format(num))
            if not flast:
                flast = 'F24'
            ff.append((flast, i))
            flast = i
            sad12.append(('C22', i))
            sad13.append(('C19', i))
            print('{}  3  {:0<8.6}  {:0<8.6}  {:0<8.6}  {:<8.6}  -1.2\
            '.format(i, co[0], co[1], co[2], 10.0 + (dif / 2760.0)))
            # with free variables:
            # print('{}  3  {:0<8.6}  {:0<8.6}  {:0<8.6}  {:<8.6}  261\
            # '.format(i, co[0], co[1], co[2], 10.0*num+1+10))
        from misc import flatten
        print('SADI ' + ' '.join(flatten(sad12)))
        print('SADI ' + ' '.join(flatten(sad13)))
        print('SADI ' + ' '.join(flatten(ff)))
        d = 0
        for dif in diffden:
            d += dif / 2760.0
        # print('PART 0')
        print('AFIX 0')
        print(d)


if __name__ == '__main__':
    pass
    """

    for i in range(0, 360, 10):
        co = cf3.rotate_atom_around_bond(F1A, C22, C19, delta=i)
        #co = cf3.rotate_fluorine_atom(F1A, C22, C19, alpha=i)
        print('F{}  3  {:0<8.6}  {:0<8.6}  {:0<8.6}  10.08  -1.5'.format(i, *co))
    
    
    
    
    for i in range(0, 360, 10):
        co = cf3.rotate_atom_around_bond(F1A, C22, C19, delta=i)
        #co = cf3.rotate_fluorine_atom(F1A, C22, C19, alpha=i)
        print('F{}  3  {:0<8.6}  {:0<8.6}  {:0<8.6}  10.08  -1.5'.format(i, *co))
    """


def calc_ellipsoid_axes(coords, uvals, cell, probability=0.5, longest=True):
    """
    This method calculates the principal axes of an ellipsoid as list of two
    fractional coordinate triples.
    Many thanks to R. W. Grosse-Kunstleve and P. D. Adams
    for their great publication on the handling of atomic anisotropic displacement
    parameters:
    R. W. Grosse-Kunstleve, P. D. Adams, J Appl Crystallogr 2002, 35, 477–480.

    F = ... * exp ( -2π²[ h²(a*)²U11 + k²(b*)²U22 + ... + 2hka*b*U12 ] )

    SHELXL atom:
    Name type  x      y      z    occ     U11 U22 U33 U23 U13 U12
    F3    4    0.210835   0.104067   0.437922  21.00000   0.07243   0.03058 =
       0.03216  -0.01057  -0.01708   0.03014
    >>> import src.mpmath as mpm
    >>> cell = [10.5086, 20.9035, 20.5072, 90, 94.13, 90]
    >>> coords = [0.210835,   0.104067,   0.437922]
    >>> uvals = [0.07243, 0.03058, 0.03216, -0.01057, -0.01708, 0.03014]
    >>> l = calc_ellipsoid_axes(coords, uvals, cell, longest=True)
    >>> print(mpm.nstr(l))
    [[0.24765096, 0.11383281, 0.43064756], [0.17401904, 0.09430119, 0.44519644]]
    >>> calc_ellipsoid_axes(coords, uvals, cell, longest=False)
    [[[0.24765096, 0.11383281, 0.43064756], [0.218406, 0.09626142, 0.43746127], [0.21924358, 0.10514684, 0.44886868]], [[0.17401904, 0.09430119, 0.44519644], [0.203264, 0.11187258, 0.43838273], [0.20242642, 0.10298716, 0.42697532]]]
    >>> cell = (10.5086, 20.9035, 20.5072, 90, 94.13, 90)
    >>> coords = [0.210835,   0.104067,   0.437922]
    >>> uvals = [0.07243, -0.03058, 0.03216, -0.01057, -0.01708, 0.03014]
    >>> calc_ellipsoid_axes(coords, uvals, cell, longest=True)
    <BLANKLINE>
    Ellipsoid is non positive definite!
    <BLANKLINE>
    False

    >>> uvals = [0.07243, 0.03058, 0.03216, -0.01057, -0.01708]
    >>> calc_ellipsoid_axes(coords, uvals, cell, longest=False)
    Traceback (most recent call last):
    ...
    Exception: 6 Uij values have to be supplied!

    >>> cell = (10.5086, 20.9035, 90, 94.13, 90)
    >>> coords = [0.210835,   0.104067,   0.437922]
    >>> uvals = [0.07243, 0.03058, 0.03216, -0.01057, -0.01708, 0.03014]
    >>> calc_ellipsoid_axes(coords, uvals, cell, longest=True)
    Traceback (most recent call last):
    ...
    Exception: cell needs six parameters!

    :param coords: coordinates of the respective atom in fractional coordinates
    :type coords: list
    :param uvals: Uij valiues of the respective ellipsoid on fractional
                  basis like in cif and SHELXL format
    :type uvals: list
    :param cell: unit cell of the structure: a, b, c, alpha, beta, gamma
    :type cell:  list
    :param probability: thermal probability of the ellipsoid
    :type probability: float or int
    :param longest: not always the length is important. make to False to
                    get all three coordiantes of the ellipsoid axes.
    :type longest: boolean

    """
    probability += 1
    # Uij is symmetric:
    if len(uvals) != 6:
        raise ValueError('6 Uij values have to be supplied!')
    if len(cell) != 6:
        raise ValueError('cell needs six parameters!')
    # orthogonalization matrix that transforms the fractional coordinates
    # with respect to a crystallographic basis system to coordinates
    # with respect to a Cartesian basis:
    _A = A(cell).orthogonal_matrix
    Ucart = ufrac_to_ucart(_A, cell, uvals)
    # print(Ucart)
    # E => eigenvalues, Q => eigenvectors:
    E, Q = mpm.eig(mpm.matrix(Ucart))
    # calculate vectors of ellipsoid axes
    try:
        sqrt(E[0])
        sqrt(E[1])
        sqrt(E[2])
    except ValueError:
        print('\nEllipsoid is non positive definite!\n')
        return False
    v1 = mpm.matrix([Q[0, 0], Q[1, 0], Q[2, 0]])
    v2 = mpm.matrix([Q[0, 1], Q[1, 1], Q[2, 1]])
    v3 = mpm.matrix([Q[0, 2], Q[1, 2], Q[2, 2]])
    v1i = v1 * (-1)
    v2i = v2 * (-1)
    v3i = v3 * (-1)
    # multiply probability (usually 50%)
    e1 = sqrt(E[0]) * probability
    e2 = sqrt(E[1]) * probability
    e3 = sqrt(E[2]) * probability
    # scale axis vectors to eigenvalues
    v1, v2, v3, v1i, v2i, v3i = v1 * e1, v2 * e2, v3 * e3, v1i * e1, v2i * e2, v3i * e3
    # find out which vector is the longest:
    length = mpm.norm(v1)
    v = 0
    if mpm.norm(v2) > length:
        length = mpm.norm(v2)
        v = 1
    elif mpm.norm(v3) > length:
        length = mpm.norm(v3)
        v = 2
    # move vectors back to atomic position
    atom = mpm.matrix(_A) * mpm.matrix(coords)
    v1, v1i = v1 + atom, v1i + atom
    v2, v2i = v2 + atom, v2i + atom
    v3, v3i = v3 + atom, v3i + atom
    # go back into fractional coordinates:
    a1 = cart_to_frac(v1, cell)
    a2 = cart_to_frac(v2, cell)
    a3 = cart_to_frac(v3, cell)
    a1i = cart_to_frac(v1i, cell)
    a2i = cart_to_frac(v2i, cell)
    a3i = cart_to_frac(v3i, cell)
    allvec = [[a1, a2, a3], [a1i, a2i, a3i]]
    if longest:
        # only the longest vector
        return [allvec[0][v], allvec[1][v]]
    else:
        # all vectors:
        return allvec



# /usr/bin/env python
# -*- encoding: utf-8 -*-
# möp
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42):
# <dkratzert@gmx.de> wrote this file. As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some day,
# and you think this stuff is worth it, you can buy me a beer in return.
# Daniel Kratzert
# ----------------------------------------------------------------------------
#

import os
import sys
from datetime import datetime
from pathlib import Path

import atomhandling
from afix import remove_duplicate_restraints, write_dbhead_to_file
from atomhandling import Elem_2_Sfac, rename_restraints_atoms
from constants import isoatomstr, sep_line, width
from dbfile import ImportGRADE, print_search_results, ParseDB, search_fragment_name
from dsrparse import DSRParser
from fit.quatfit import centroid, fit_fragment
from misc import cart_to_frac, chunks, frac_to_cart, touch, wrap_headlines, matrix_plus_vect, subtract_vect
from options import OptionsParser
from refine import ShelxlRefine
from resfile import ResList, ResListEdit, filename_wo_ending
from resi import Resi, remove_resi
from restraints import Restraints
from terminalsize import get_terminal_size
from version import VERSION
import time

# dont forget to change version in Innoscript file, spec file and deb file.
minuse = ((width // 2) - 7) * '-'
program_name = '\n{} D S R - v{} {}'.format(minuse, VERSION, minuse)

# TODO and ideas:
"""
- make DSR Python3-only
- Split mode
  - calc principal axis for each ellipsoid
  - generate atoms
  - decide which new bond to new and old atoms would leave the bond length intact
  - generate parts and restraints
- Add Rcomplete

"""


class DSR(object):
    """
    main class
    """

    def __init__(self, options):
        """
        """
        time1 = time.perf_counter()
        # options from the commandline options parser:
        self.options = options
        self.external = False
        self.fragment = ''
        self.helpmsg = "*** Please ask dkratzert@gmx.de for help ***"
        self.res_file = self.options.res_file
        if self.options.external_restr:
            self.external = True
            self.res_file = self.options.external_restr
        self.export_fragment = self.options.export_fragment
        if self.export_fragment:
            self.res_file = self.options.export_fragment
        if self.export_fragment:
            self.fragment = self.export_fragment
        self.export_clip = self.options.export_clip
        if self.export_clip:
            self.fragment = self.export_clip
        self.import_grade = self.options.import_grade
        self.export_all = self.options.export_all
        self.list_db = self.options.list_db
        self.list_db_csv = self.options.list_db_csv
        self.no_refine = self.options.no_refine
        self.invert = self.options.invert
        self.rigid = self.options.rigid_group
        self.search_string = self.options.search_string
        self.search_extern = self.options.search_extern
        self.head_csv = self.options.head_for_gui
        if self.head_csv:
            self.fragment = self.head_csv
        if self.options.selfupdate:
            import selfupdate
            selfupdate.update_dsr()
            sys.exit()
        #################################
        self.maindb_path = ''
        self.userdb_path = ''
        try:
            self.set_database_locations()
            self.gdb = ParseDB(self.maindb_path, self.userdb_path)
        except Exception as e:  # @UnusedVariable
            print("*** Initializing the database failed ***")
            raise
        #  List of database Fragments:
        if self.list_db_csv:
            print('DSR version: {}'.format(VERSION))
            for i in self.gdb.list_fragments():
                print('{};;{};;{};;{}'.format(i[0], i[3], i[1], i[2]))
            sys.exit()
        try:
            from export import Export
            self.export = Export(gdb=self.gdb, invert=self.invert)
        except Exception as e:
            print("*** Unable to export informations from DSR ***")
            print(e)
            raise
            sys.exit()
        #################################
        if not any(list(vars(self.options.all_options).values()) + [self.res_file]):
            self.options.error()
        if self.head_csv:
            self.head_to_gui()
        if self.search_extern:
            result = search_fragment_name(self.search_extern, self.gdb, numresults=8)
            for i in result:
                print('{};;{};;{};;{}'.format(i[0], i[1], i[2], i[3]))
            sys.exit()
        print(program_name)
        if self.list_db:
            self.list_dbentries()
        if self.search_string:
            result = search_fragment_name(self.search_string, self.gdb, numresults=7)
            print_search_results(result)
            sys.exit()
        # Export !all! fragments
        if self.export_all:
            self.export.export_all_fragments()
        # Export one fragment
        if self.export_fragment:
            print(f'Exporting "{self.fragment}" to {self.fragment}.res')
            self.fragment = self.export_fragment
            self.export.write_res_file(self.fragment)
            sys.exit()
        if self.export_clip:
            self.export.export_to_clip(self.fragment)
            sys.exit()
        # Import a GRADE fragment
        if self.import_grade:
            mog = ImportGRADE(self.import_grade, self.gdb, self.invert, self.maindb_path, self.userdb_path)
            mog.write_user_database()
            sys.exit()
        if not self.res_file:
            self.options.error()
        self.rl = ResList(self.res_file)
        self.reslist = self.rl.get_res_list()
        self.main()
        time2 = time.perf_counter()
        runtime = (time2 - time1)
        print('Runtime: {:>.1f} s'.format(runtime))
        print('DSR run complete.')

    ###############################################################################

    def set_database_locations(self):
        homedir = os.path.expanduser("~")
        self.userdb_path = os.path.join(homedir, "dsr_user_db.txt")
        if not os.path.isfile(self.userdb_path):
            touch(self.userdb_path)
        main_dbdir = os.environ.get("DSR_DIR", None)
        if main_dbdir:
            self.maindb_path = os.path.join(main_dbdir, 'dsr_db.txt')
        else:
            self.maindb_path = str(Path(__file__).parent / 'dsr_db.txt')
        return True

    def head_to_gui(self):
        """
        Exports current fragment header and atoms to the GUI
        """
        atoms = []
        try:
            atoms = self.export.export_to_gui(self.fragment)
        except Exception as e:
            # print(e)
            print("*** Could not get atom information ***")
            print(self.helpmsg)
            sys.exit()
        print("\n<atoms>")
        print(atoms)
        print("</atoms>")
        # prints most of the needed info:
        self.gdb.get_head_for_gui(self.fragment)
        sys.exit()

    def list_dbentries(self):
        """
        list all entries in the db.
        """
        dbdir = os.path.expanduser('~')
        fragnames = []
        num = 0
        try:
            (width, _) = get_terminal_size()
        except():
            width = 80
        print('\n Entries found in the databases:\n')
        print(' Fragment         | Line | DB Name    | Full name, Comments ')
        print(sep_line)
        for num, line in enumerate(self.gdb.list_fragments()):
            fragnames.append(line[0])
            line = ' {:<17}| {:<5}| {:<11}| {}'.format(*line)
            print(line[:width - 1])
        print('\n {} Fragments in the database(s).'.format(num),
              '\n Feel free to add more fragments to "{}dsr_user_db.txt"\n'.format(dbdir + os.path.sep))
        for fragment in fragnames:
            self.gdb.check_consistency(fragment)
            self.gdb.check_db_atom_consistency(fragment)
            self.gdb.check_db_restraints_consistency(fragment)
            self.gdb.check_sadi_consistence(fragment)
        from selfupdate import is_update_needed
        if is_update_needed(silent=True):
            print("\n*** An update for DSR is available. You can update with 'dsr -u' ***")
        sys.exit()

    def main(self):
        """
        main object to run DSR as command line program
        """
        dbatoms = []
        # The database content:
        basefilename = filename_wo_ending(self.res_file)
        if not basefilename:
            print('*** Illegal option ***')
            sys.exit()
        if len(self.reslist) == 0:
            print("*** The input file is empty. Can not proceed! ***")
            sys.exit()
        find_atoms = atomhandling.FindAtoms(self.reslist)
        rle = ResListEdit(self.reslist, find_atoms)
        dsrp = DSRParser(self.reslist)
        self.fragment = dsrp.fragment
        restraints = self.gdb.get_restraints(self.fragment)  # this is only executed once
        db_residue_string = self.gdb.get_resi(self.fragment)
        dbatoms = self.gdb.get_atoms(self.fragment, self.invert)  # only the atoms of the dbentry as list
        # the atomtypes of the dbentry as list e.g. ['C', 'N', ...]
        db_atom_types = atomhandling.get_atomtypes(dbatoms)
        sf = atomhandling.SfacTable(self.reslist, db_atom_types)
        sfac_table = sf.set_sfac_table()  # from now on this sfac table is set
        resi = Resi(dsrp, db_residue_string, find_atoms)
        # line where the dsr command is found in the resfile:
        if dsrp.cf3_active:
            from cf3fit import CF3
            cf3 = CF3(rle, find_atoms, self.reslist, self.fragment, sfac_table,
                      basefilename, dsrp, resi, self.res_file, self.options)
            if self.fragment == 'cf3':
                cf3.cf3(afix='130')
            if self.fragment == 'cf6':
                cf3.cf3(afix='120')
            if self.fragment == 'cf9':
                cf3.cf9()
            print('\nFinished...')
            sys.exit()
        # checks have to be after CF3, CF6 etc.
        self.gdb.check_consistency(self.fragment)
        self.gdb.check_db_atom_consistency(self.fragment)
        self.gdb.check_db_restraints_consistency(self.fragment)
        self.gdb.check_sadi_consistence(self.fragment)
        if dsrp.occupancy:
            rle.set_free_variables(dsrp.occupancy)
        restraints = remove_resi(restraints)
        # corrects the atom type according to the previous defined global sfac table:
        dbatoms = atomhandling.set_final_db_sfac_types(db_atom_types, dbatoms, sfac_table)
        if not dsrp.unit_line:
            print('*** No UNIT instruction in res file found! Can not proceed! ***')
        print('Inserting {} into res File.'.format(self.fragment))
        if self.invert:
            print('Fragment inverted.')
        print('Source atoms: {}'.format(', '.join(dsrp.source)))
        print('Target atoms: {}'.format(', '.join(dsrp.target)))
        shx = ShelxlRefine(self.reslist, basefilename, find_atoms, self.options)
        shx.backup_shx_file()
        # several checks if the atoms in the dsr command line are consistent
        atomhandling.check_source_target(dsrp.source, dsrp.target, dbatoms)
        num = atomhandling.NumberScheme(self.reslist, dbatoms, dsrp)
        # returns also the atom names if residue is active
        fragment_numberscheme = num.get_fragment_number_scheme()
        print('Fragment atom names: {}'.format(', '.join(fragment_numberscheme)))
        dfix_head = ''
        if dsrp.dfix:
            restr = Restraints(self.fragment, self.gdb)
            dfix_12 = restr.get_formated_12_dfixes()
            dfix_13 = restr.get_formated_13_dfixes()
            flats = restr.get_formated_flats()
            restraints = dfix_12 + dfix_13 + flats
        # ##########Not using SHELXL for fragment fit: ###########

        print("--- Using fast fragment fit ---")
        if self.options.target_coords:
            target_coords = chunks(self.options.target_coords, 3)
        else:
            # {'C1': ['1.123', '0.7456', '3.245']}
            target_coordinates = find_atoms.get_atomcoordinates(dsrp.target)
            target_coords = [target_coordinates[key] for key in dsrp.target]
        # Uppercase is important here to avoid KeyErrors in source_atoms generation
        atnames = self.gdb.get_atomnames(self.fragment, uppercase=True)
        source_atoms = dict(zip(atnames, self.gdb.get_coordinates(self.fragment, cartesian=True, invert=self.invert)))
        # Coordinates only from the source, not the entire fragment:
        source_coords = [source_atoms[x] for x in dsrp.source]
        target_coords = [frac_to_cart(x, rle.cell) for x in target_coords]
        # The source and target atom coordinates are fitted first. Then The complete fragment
        # is rotated and translated to the target position as calculated before.
        # parameter cartiesian has to be false here:
        fragment_coords = self.gdb.get_coordinates(self.fragment, cartesian=True, invert=self.invert)
        fitted_fragment, rmsd = fit_fragment(fragment_coords,
                                             source_atoms=source_coords,
                                             target_atoms=target_coords)
        # Moving back to the position of the first atom to have a reference:
        # I have to make sure that I use the centroid of the correct atoms from target and source,
        # otherwise the fragment is shifted to a wrong position.
        # The third atom from the fragment e.g. has to be the third from the fragment to get
        # the correct centroid:
        center_difference = subtract_vect(centroid(target_coords),
                                          centroid([list(fitted_fragment)[atnames.index(dsrp.source[x])] for x in
                                                    range(len(source_coords))]))
        # finishing shift to correct centroid:
        fitted_fragment = matrix_plus_vect(fitted_fragment, center_difference)
        # Or even lower than 0.1?
        if rmsd < 0.1:
            print('Fragment fit successful with RMSD of: {:8.3}'.format(rmsd))
        else:
            print('*** Fragment fit might have failed with RMSD of: {:8.3} ***'.format(rmsd))
        fitted_fragment = [cart_to_frac(x, rle.cell) for x in fitted_fragment]
        afix_entry = []
        e2s = Elem_2_Sfac(sfac_table)
        for at, coord, atype in zip(fragment_numberscheme, fitted_fragment, db_atom_types):
            sfac_num = str(e2s.elem_2_sfac(atype))
            if dsrp.occupancy:
                occ = float(dsrp.occupancy)
            else:
                occ = 11.0
            afix_entry.append(isoatomstr.format(at, sfac_num, coord[0], coord[1], coord[2], occ, 0.03))
        afix_entry = "\n".join(afix_entry)
        new_atomnames = list(reversed(fragment_numberscheme))
        same_resi = ''
        if not dsrp.resiflag:
            restraints = rename_restraints_atoms(new_atomnames, self.gdb.get_atomnames(self.fragment), restraints)
        else:
            restraints = resi.format_restraints(restraints)
            # SADI\n
            same_resi = ["SAME_{} {} > {}\n".format(resi.get_residue_class, new_atomnames[-1], new_atomnames[0])]
        # Adds a "SAME_resiclass firstatom > lastatom" to the afix:
        if not self.options.rigid_group:
            restraints += same_resi
            # if dsrp.resiflag:  # <- Or should I do this?
            restraints += ["SIMU 0.04 0.08 1"]
        if not self.options.external_restr:
            restraints = remove_duplicate_restraints(self.reslist, restraints, resi.get_residue_class)
        restraints = wrap_headlines(restraints)
        dfx_file_name = ''
        if dsrp.part:
            afix_entry = "PART {}  {}\n".format(dsrp.part, dsrp.occupancy) + afix_entry + "\nPART 0"
        if dsrp.resiflag:
            afix_entry = 'RESI {} {}\n{}\nRESI 0'.format(resi.get_residue_class, resi.get_resinumber, afix_entry)
        if self.options.rigid_group:
            afix_entry = 'AFIX 9\n' + afix_entry
        if self.options.external_restr and not self.rigid:
            pname, ext = os.path.splitext(basefilename + '.dfix')
            if dsrp.dfix:
                dfx_file_name = pname + "_dfx" + ext
            else:
                dfx_file_name = pname + ext
            dfx_file_name = write_dbhead_to_file(dsrp, dfx_file_name, restraints, resi.get_residue_class,
                                                 resi.get_resinumber)
            if dsrp.resiflag:
                restraints = 'REM Restraints for residue {}:\n+{}\n'.format(resi.get_residue_class, dfx_file_name)
            else:
                restraints = 'REM Restraints for DSR fragment:\n+{}\n'.format(dfx_file_name)
        if self.options.rigid_group:
            afix_entry += '\nAFIX 0\n'

        # Adds the origin of restraints and fragment to res file:
        import textwrap
        source = textwrap.wrap("REM Restraints for Fragment {}, {} from: {}. "
                               "Please cite https://doi.org/10.1107/S1600576718004508".format(
            self.fragment,
            self.gdb.get_fragment_name(self.fragment),
            self.gdb.get_src(self.fragment)),
            width=74, subsequent_indent='REM ')
        if dsrp.resi:
            hfixes = '\n'.join(self.gdb.get_hfixes(self.fragment, resi.get_residue_class))
            if hfixes:
                hfixes += '\n'
        else:
            hfixes = ''
        source = '\n'.join(source) + '\n' + hfixes
        # check if restraints already inserted:
        for line in self.reslist:
            try:
                if line.split()[4] == self.fragment + ',':
                    source = ''
                    break
            except IndexError:
                continue
        # + 'AFIX 0\n' before hklf seems to be not needed after shelx-2013:
        self.reslist[dsrp.hklf_line - 1] = self.reslist[dsrp.hklf_line - 1] + afix_entry + '\n'
        if not self.rigid:
            self.reslist[dsrp.unit_line] = self.reslist[dsrp.unit_line] + source + ''.join(restraints)

        # write to file:
        self.rl.write_resfile(self.reslist, '.res')
        if dsrp.command == 'REPLACE':
            print("Replace mode active\n")
            self.rl = ResList(self.res_file)
            reslist = self.rl.get_res_list()
            self.reslist, find_atoms = atomhandling.replace_after_fit(self.rl, reslist, resi,
                                                                      fragment_numberscheme, rle.get_cell())
            self.rl.write_resfile(self.reslist, '.res')
        os.remove(shx.backup_file)


class Multilog(object):
    """
    This class copies all output from stdout and stderr to a file
    It acts like tee with following usage:
    sys.stdout = multifile([sys.stdout, lstfileobj])
    """

    def __init__(self, files):
        self._files = files

    def __getattr__(self, attr, *args):
        return self._wrap(attr, *args)

    def _wrap(self, attr, *args):
        def g(*a, **kw):
            res = ''
            for f in self._files:
                res = getattr(f, attr, *args)(*a, **kw)
            return res

        return g


def main():
    options = None
    lstfile = ''
    is_listfile = False
    try:
        lstfile = open('./dsr-log.lst', 'w')
    except IOError:
        pass
    else:
        sys.stdout = Multilog([sys.stdout, lstfile])
        sys.stderr = Multilog([sys.stderr, lstfile])
        is_listfile = True
    try:
        options = OptionsParser(program_name)
        if is_listfile:
            lstfile.write('Python version: {}\n'.format(sys.version))
            lstfile.write("Date: {} \n".format(datetime.now().strftime("%Y-%m-%d %H:%M:%S")))
            lstfile.write(options.__str__())
        DSR(options)
    except Exception:
        import platform

        if is_listfile:
            lstpath = os.path.abspath(lstfile.name)
            lst = 'the file "{}" \nand '.format(lstpath)
        else:
            lst = "this error message and "
        print('\n*** Congratulations! You found a bug in DSR. Please send {}the .res file '
              '(if possible) to dkratzert@gmx.de ***\n\n'.format(lst))
        print('DSR version: {}'.format(VERSION))
        print('Python version: {}'.format(sys.version))
        if options:
            print('Commandline: {}'.format(options.all_options))
        print('Platform: {} {}, {}'.format(platform.system(),
                                           platform.release(), ' '.join(platform.uname())))
        print(datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
        raise


if __name__ == '__main__':
    main()



# -*- encoding: utf-8 -*-
# möp
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42):
# <dkratzert@gmx.de> wrote this file. As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some day,
# and you think this stuff is worth it, you can buy me a beer in return.
# Daniel Kratzert
# ----------------------------------------------------------------------------
#
# each residue number must always have the same class!
# e.g 1 can be only TOL, not also NET4 but Tol can be 1 or 2.
#
# Wenn das dict leer ist werden überhaupt keine RESI betreffenden dinge in das insfile übernommen



import sys

from atomhandling import FindAtoms
from constants import RESTRAINT_CARDS
from dsrparse import DSRParser


def remove_resi(head):
    # type: (list) -> list
    """
    removes all resi commands and classes from restraints
    :param head: database header of the current fragment
    :type head: list
    """
    rhead = []  # head without resi
    delhead = []
    for dummy, line in enumerate(head):
        line = line.split()
        try:
            if line[0][:4] in RESTRAINT_CARDS:
                line[0] = line[0].split('_')[0]
        except IndexError:
            continue
        line = ' '.join(line)
        delhead.append(line)
    for line in delhead:
        line = line.strip(' \n\r').upper()
        if line.startswith('RESI'):
            continue
        rhead.append(line)
    return rhead


class Resi(object):
    def __init__(self, dsrp, db_residue_string, find_atoms):
        # type: (DSRParser, str, FindAtoms) -> NotImplemented
        """
        Handles the RESI instructions and restraints
        :param dsrp: Dsrparse object
        :type dsrp: object
        :param reslist: res file as list
        :type reslist: list
        :param dbhead: database header with restraints and residue
        :type dbhead: list
        :param db_residue_string: db residue string
        :type db_residue_string: string
        :param find_atoms: find_atoms object
        :type find_atoms: object
        """
        self._atoms_in_reslist = find_atoms.atoms_as_residues()
        self._residues_in_res = sorted(list(self._atoms_in_reslist))
        self.dsrp = dsrp
        self._resi_dict_dsr_command = {'class': None, 'number': None, 'alias': None}
        if self.dsrp.resiflag:
            # use a residue
            if self.dsrp.resi:
                # a residue was defined (class, number, or both)
                self._resi_dict_dsr_command = self.get_resi_syntax(self.dsrp.resi)
            else:
                # no residue defined, only active (use class from DB)
                self._resi_dict_dsr_command = {'class': None, 'number': None, 'alias': None}
        try:
            # use residue class from db if not given in dsr command:
            self._db_resi_list = db_residue_string.split()
        except AttributeError:
            print('No valid residue "RESI classname" in the database entry '
                  'of {} found.'.format(self.dsrp.fragment))
            sys.exit()
        self._resi_dict_db = self.get_resi_syntax(self._db_resi_list)
        self._combined_resi = self.build_up_residue()

    @property
    def get_resinumber(self):
        """
        Returns the residue number of the currently fitted fragment
        :rtype self._combined_resi['number']: string
        """
        return self._combined_resi['number']

    @property
    def get_residue_class(self):
        """
        Returns the residue class of the currently fitted fragment. Also is an indicator 
        if residues are active.
        :rtype self._combined_resi['class']: str
        """
        return self._combined_resi['class']

    def format_restraints(self, head):
        """
        in case of RESI, format the restraints like "SAME_class"
        """
        newhead = []
        for line in head:
            line = line.upper()
            line = line.split()
            try:
                line[0]
            except:
                continue
            if line[0] in RESTRAINT_CARDS:
                line[0] = line[0] + '_' + self._combined_resi['class']
                line = ' '.join(line)
            else:
                line = ' '.join(line)
            newhead.append(line)
        return newhead

    def get_unique_resinumber(self, resinum):
        # type: (str) -> str
        """
        Finds a unique resi number. If the number is already unique
        the given is used.
        :param resinum: residue number of the fragment
        :type resinum: string
        :return resinum: unique residue number
        """
        new_num = '1'
        if resinum == '0':
            while new_num in self._residues_in_res:
                new_num = str(int(new_num) + 1)
            return new_num
        elif resinum in self._residues_in_res and not resinum == '0':
            print('Warning: The residue number "{}" you have chosen is already ' \
                  'in use!'.format(resinum))
            while new_num in self._residues_in_res:
                new_num = str(int(new_num) + 1)
            return new_num
        else:
            return resinum

    def build_up_residue(self):
        # type: () -> dict
        """
        Decides which class and residue number should be used for the fragment.
        Returns a final dict with the residue settings.
        self._resi_dict_dsr_command is False if resi is enabled but no values given.
        """
        final_residue = {'class': None, 'number': None, 'alias': None}
        resiclass = None
        resinum = '0'
        resialias = None

        #### for the db entry
        if self._resi_dict_db['alias']:
            resialias = self._resi_dict_db['alias']
        if self._resi_dict_db['class']:
            resiclass = self._resi_dict_db['class'].upper()
        if self._resi_dict_db['number']:
            resinum = self._resi_dict_db['number']
        #### for the comlist entry
        if not self._resi_dict_dsr_command:
            return final_residue
        else:
            if self._resi_dict_dsr_command['alias']:
                resialias = self._resi_dict_dsr_command['alias']
            if self._resi_dict_dsr_command['class']:
                resiclass = self._resi_dict_dsr_command['class'].upper()
            if self._resi_dict_dsr_command['number']:
                resinum = self._resi_dict_dsr_command['number']
        final_residue = {'class': resiclass, 'number': resinum, 'alias': resialias}
        if resinum == '0':
            final_residue['number'] = self.get_unique_resinumber(resinum)
            print('No residue number was given. Using residue number {}.'.format(final_residue['number']))
        else:
            final_residue['number'] = self.get_unique_resinumber(resinum)
            print('Using residue number {}.'.format(final_residue['number']))
        if final_residue['class']:
            return final_residue

    @staticmethod
    def _wrong_syntax():
        print('This is not a valid RESIdue syntax!')

    def get_resi_syntax(self, inresi):
        """
        Checks if resi class, number and/or alias are present and valid.
        start with digit-> rest auch digit-> resinumber or alias
        start with letter-> rest letter or digit -> residue class

        The return value of this method is {'class', 'number', 'alias'}
        as a dictionary. Alias is empty if not given.

        The return value of just "RESI" in the command line is an empty dict
        :param resi: residue definition like ['3', 'CF3']
        :type resi: list
        """
        resi = inresi[:]
        resi_dict = {
            'class': '',
            'number': '',
            'alias': ''}
        if not resi:
            print('No valid RESI instruction found in the database entry!')
            sys.exit()
        try:
            resi.sort()
        except AttributeError:
            return resi_dict
        # any character of class must be a letter (New syntax in SHELXL 2017/1):
        if any([x.isalpha() for x in resi[-1]]):
            resi_dict['class'] = resi.pop()
        if len(resi) > 0:
            if str.isdigit(resi[0][0]):  # first character of number must be an digit
                resi_dict['number'] = resi[0]
                del resi[0]
            else:
                self._wrong_syntax()
                # not totally correct, but less difficult:
                print('Only digits are allowed for the residue number.')
                sys.exit()
        if len(resi) > 0:
            if str.isdigit(resi[0]):
                resi_dict['alias'] = resi[0]
                del resi[0]
            else:
                self._wrong_syntax()
                print('Only digits are allowed in the residue alias.')
                sys.exit()
        if resi_dict['number']:
            if len(resi_dict['number']) > 4:
                self._wrong_syntax()
                print('Only four digits allowed in residue number!')
                sys.exit()
        return resi_dict


if __name__ == '__main__':
    pass
